# Copyright 2019 The TensorFlow Probability Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""SoftClip bijector."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Dependency imports

import tensorflow.compat.v2 as tf

from tensorflow_probability.python import util as tfp_util

from tensorflow_probability.python.bijectors import bijector
from tensorflow_probability.python.bijectors import chain
from tensorflow_probability.python.bijectors import scale
from tensorflow_probability.python.bijectors import shift
from tensorflow_probability.python.bijectors import softplus

from tensorflow_probability.python.internal import assert_util
from tensorflow_probability.python.internal import auto_composite_tensor
from tensorflow_probability.python.internal import dtype_util
from tensorflow_probability.python.internal import parameter_properties
from tensorflow_probability.python.internal import tensor_util


__all__ = [
    'SoftClip',
]


# @auto_composite_tensor.auto_composite_tensor(
#     omit_kwargs=('name',), module_name='tfp.bijectors')
class SoftClip(bijector.Bijector):
  """Bijector that approximates clipping as a continuous, differentiable map.

  The `forward` method takes unconstrained scalar `x` to a value `y` in
  `[low, high]`. For values within the interval and far from the bounds
  (`low << x << high`), this mapping is approximately the identity mapping.

  ```python
  b = tfb.SoftClip(low=-10., high=10.)
  b.forward([-15., -7., 1., 9., 20.])
    # => [-9.993284, -6.951412,  0.9998932,  8.686738,  9.999954 ]
  ```

  The softness of the clipping can be adjusted via the `hinge_softness`
  parameter. A sharp constraint (`hinge_softness < 1.0`) will approximate
  the identity mapping very well across almost all of its range, but may
  be numerically ill-conditioned at the boundaries. A soft constraint
  (`hinge_softness > 1.0`) corresponds to a smoother, better-conditioned
  mapping, but creates a larger distortion of its inputs.

  ```python
  b_hard = SoftClip(low=-5, high=5., hinge_softness=0.1)
  b_soft.forward([-15., -7., 1., 9., 20.])
    # => [-10., -7., 1., 8.999995,  10.]

  b_soft = SoftClip(low=-5, high=5., hinge_softness=10.0)
  b_soft.forward([-15., -7., 1., 9., 20.])
    # => [-6.1985435, -3.369276,  0.16719627,  3.6655345,  7.1750355]
  ```

  Note that the outputs are always in the interval `[low, high]`, regardless
  of the `hinge_softness`.

  #### Example use

  A trivial application of this bijector is to constrain the values sampled
  from a distribution:

  ```python
  dist = tfd.TransformedDistribution(
    distribution=tfd.Normal(loc=0., scale=1.),
    bijector=tfb.SoftClip(low=-5., high=5.))
  samples = dist.sample(100)  # => samples guaranteed in [-10., 10.]
  ```

  A more useful application is to constrain the values considered
  during inference, preventing an inference algorithm from proposing values
  that cause numerical issues. For example, this model will return a `log_prob`
  of `NaN` when `z` is outside of the range `[-5., 5.]`:

  ```python
  dist = tfd.JointDistributionNamed({
    'z': tfd.Normal(0., 1.0)
    'x': lambda z: tfd.Normal(
                     loc=tf.log(25 - z**2), # Breaks if z >= 5 or z <= -5.
                     scale=1.)})
  ```

  Using SoftClip allows us to keep an inference algorithm in the feasible
  region without distorting the inference geometry by very much:

  ```python
  target_log_prob_fn = lambda z: dist.log_prob(z=z, x=3.)  # Condition on x==3.

  # Use SoftClip to ensure sampler stays within the numerically valid region.
  mcmc_samples = tfp.mcmc.sample_chain(
    kernel=tfp.mcmc.TransformedTransitionKernel(
      tfp.mcmc.HamiltonianMonteCarlo(
        target_log_prob_fn=target_log_prob_fn,
        num_leapfrog_steps=2,
        step_size=0.1),
      bijector=tfb.SoftClip(-5., 5.)),
    trace_fn=None,
    current_state=0.,
    num_results=100)
  ```

  #### Mathematical Details

  The constraint is built by using `softplus(x) = log(1 + exp(x))` as a smooth
  approximation to `max(x, 0)`. In combination with affine transformations, this
  can implement a constraint to any scalar interval.

  In particular, translating `softplus` gives a generic lower bound constraint:

  ```
  max(x, low) =  max(x - low, 0) + low
              ~= softplus(x - low) + low
              := softlower(x)
  ```

  Note that this quantity is always greater than `low` because `softplus` is
  positive-valued. We can also implement a soft upper bound:

  ```
  min(x, high) =  min(x - high, 0) + high
               = -max(high - x, 0) + high
              ~= -softplus(high - x) + high
              := softupper(x)
  ```

  which, similarly, is always less than `high`.

  Composing these bounds as `softupper(softlower(x))` gives a quantity bounded
  above by `high`, and bounded below by `softupper(low)` (because `softupper`
  is monotonic and its input is bounded below by `low`). In general, we will
  have `softupper(low) < low`, so we need to shrink the interval slightly
  (by `(high - low) / (high - softupper(low))`) to preserve the lower bound.
  The two-sided constraint is therefore:

  ```python
  softclip(x) := (softupper(softlower(x)) - high) *
                   (high - low) / (high - softupper(low)) + high
               = -softplus(high - low - softplus(x - low)) *
                   (high - low) / (softplus(high-low)) + high
  ```

  Due to this rescaling, the bijector can be mildly asymmetric. Values
  of equal distance from the endpoints are mapped to values with slightly
  unequal distance from the endpoints; for example,

  ```python
  b = SoftConstrain(-1., 1.)
  b.forward([-0.5., 0.5.])
    # => [-0.2527727 ,  0.19739306]
  ```

  The degree of the asymmetry is proportional to the size of the rescaling
  correction, i.e., the extent to which `softupper` fails to be the identity
  map at the lower end of the interval. This is maximized when the upper and
  lower bounds are very close together relative to the hinge softness, as in
  the example above. Conversely, when the interval is wide, the required
  correction and asymmetry are very small.

  """

  def __init__(self,
               low=None,
               high=None,
               hinge_softness=None,
               validate_args=False,
               name='soft_clip'):
    """Instantiates the SoftClip bijector.

    Args:
      low: Optional float `Tensor` lower bound. If `None`, the lower-bound
        constraint is omitted.
        Default value: `None`.
      high: Optional float `Tensor` upper bound. If `None`, the upper-bound
        constraint is omitted.
        Default value: `None`.
      hinge_softness: Optional nonzero float `Tensor`. Controls the softness
        of the constraint at the boundaries; values outside of the constraint
        set are mapped into intervals of width approximately
        `log(2) * hinge_softness` on the interior of each boundary. High
        softness reserves more space for values outside of the constraint set,
        leading to greater distortion of inputs *within* the constraint set,
        but improved numerical stability near the boundaries.
        Default value: `None` (`1.0`).
      validate_args: Python `bool` indicating whether arguments should be
        checked for correctness.
      name: Python `str` name given to ops managed by this object.
    """
    parameters = dict(locals())
    with tf.name_scope(name):
      dtype = dtype_util.common_dtype(
          [low, high, hinge_softness], dtype_hint=tf.float32)
      low = tensor_util.convert_nonref_to_tensor(
          low, name='low', dtype=dtype)
      high = tensor_util.convert_nonref_to_tensor(
          high, name='high', dtype=dtype)
      hinge_softness = tensor_util.convert_nonref_to_tensor(
          hinge_softness, name='hinge_softness', dtype=dtype)

      softplus_bijector = softplus.Softplus(hinge_softness=hinge_softness)
      negate = tf.convert_to_tensor(-1., dtype=dtype)

      components = []
      if low is not None and high is not None:
        # Support reference tensors (eg Variables) for `high` and `low` by
        # deferring all computation on them until needed.
        width = tfp_util.DeferredTensor(
            pretransformed_input=high, transform_fn=lambda high: high - low)
        negated_shrinkage_factor = tfp_util.DeferredTensor(
            pretransformed_input=width,
            transform_fn=lambda w: tf.cast(  # pylint: disable=g-long-lambda
                negate * w / softplus_bijector.forward(w), dtype=dtype))

        # Implement the soft constraint from 'Mathematical Details' above:
        #  softclip(x) := -softplus(width - softplus(x - low)) *
        #                        (width) / (softplus(width)) + high
        components = [
            shift.Shift(high),
            scale.Scale(negated_shrinkage_factor),
            softplus_bijector,
            shift.Shift(width),
            scale.Scale(negate),
            softplus_bijector,
            shift.Shift(tfp_util.DeferredTensor(low, lambda x: -x))]
      elif low is not None:
        # Implement a soft lower bound:
        #  softlower(x) := softplus(x - low) + low
        components = [
            shift.Shift(low),
            softplus_bijector,
            shift.Shift(tfp_util.DeferredTensor(low, lambda x: -x))]
      elif high is not None:
        # Implement a soft upper bound:
        #  softupper(x) := -softplus(high - x) + high
        components = [shift.Shift(high),
                      scale.Scale(negate),
                      softplus_bijector,
                      scale.Scale(negate),
                      shift.Shift(high)]

      self._low = low
      self._high = high
      self._hinge_softness = hinge_softness
      self._chain = chain.Chain(components, validate_args=validate_args)

    super(SoftClip, self).__init__(
        forward_min_event_ndims=0,
        dtype=dtype,
        validate_args=validate_args,
        parameters=parameters,
        is_constant_jacobian=not components,
        name=name)

  @classmethod
  def _parameter_properties(cls, dtype):
    return dict(
        low=parameter_properties.ParameterProperties(),
        high=parameter_properties.ParameterProperties(
            default_constraining_bijector_fn=parameter_properties
            .BIJECTOR_NOT_IMPLEMENTED),
        hinge_softness=parameter_properties.ParameterProperties(
            default_constraining_bijector_fn=(
                lambda: softplus.Softplus(low=dtype_util.eps(dtype)))))

  @property
  def low(self):
    return self._low

  @property
  def high(self):
    return self._high

  @property
  def hinge_softness(self):
    return self._hinge_softness

  @classmethod
  def _is_increasing(cls):
    return True

  def _forward(self, x):
    return self._chain.forward(x)

  def _forward_log_det_jacobian(self, x):
    return self._chain.forward_log_det_jacobian(x, self.forward_min_event_ndims)

  def _inverse(self, y):
    with tf.control_dependencies(self._assert_valid_inverse_input(y)):
      return self._chain.inverse(y)  # pylint: disable=protected-access

  def _inverse_log_det_jacobian(self, y):
    with tf.control_dependencies(self._assert_valid_inverse_input(y)):
      return self._chain.inverse_log_det_jacobian(
          y, self.inverse_min_event_ndims)

  def _assert_valid_inverse_input(self, y):
    assertions = []
    if self.validate_args and self.low is not None:
      assertions += [assert_util.assert_greater(
          y, self.low,
          message='Input must be greater than `low`.')]
    if self.validate_args and self.high is not None:
      assertions += [assert_util.assert_less(
          y, self.high,
          message='Input must be less than `high`.')]
    return assertions

  def _parameter_control_dependencies(self, is_init):
    if not self.validate_args or self.low is None or self.high is None:
      return []
    assertions = []
    if is_init != (tensor_util.is_ref(self.low) or
                   tensor_util.is_ref(self.high)):
      assertions.append(assert_util.assert_greater(
          self.high, self.low,
          message='Argument `high` must be greater than `low`.'))
    return assertions



import numpy as np
import scipy.optimize as op
from george import kernels
import george
george.__version__


class GPRFit(object):
    def __init__(self, X, Y, yerr=None):
        """
        X : np.ndarray, shape = (N, D), N=samples, D=dimension
        Y : np.ndarray, shape = (N,)
        yerr : np.ndarray, shape = (N,)
        """
        self.X = X
        self.Y = Y
        self.yerr = yerr
        if self.yerr is None:
            print("no yerr given - setting to zero")
            self.yerr = np.zeros(shape=self.Y.shape)

    def setup_kernel(self, kernel):
        self.kernel = kernel

    def fit(self, solver=george.BasicSolver):
        self.gp = george.GP(self.kernel, solver=solver)
        self.gp.compute(self.X, self.yerr)

    def opt(self, method="L-BFGS-B"):
        # Define the objective function (negative log-likelihood in this case).
        def nll(p):
            self.gp.set_parameter_vector(p)
            ll = self.gp.log_likelihood(self.Y, quiet=True)
            return -ll if np.isfinite(ll) else 1e25

        # And the gradient of the objective function.
        def grad_nll(p):
            self.gp.set_parameter_vector(p)
            return -self.gp.grad_log_likelihood(self.Y, quiet=True)

        # You need to compute the GP once before starting the optimization.
        self.gp.compute(self.X, self.yerr)

        # Print the initial ln-likelihood.
#         print(self.gp.log_likelihood(self.Y))

        # Run the optimization routine.
        p0 = self.gp.get_parameter_vector()
        results = op.minimize(nll, p0, jac=grad_nll, method=method)

        # Update the kernel and print the final log-likelihood.
        self.gp.set_parameter_vector(results.x)
#         print(self.gp.log_likelihood(self.Y))

    def predict(self, X):
        mean, var = self.gp.predict(self.Y, X, return_var=True)
        return mean, var



from sklearn.preprocessing import StandardScaler
import numpy as np
import warnings


def make_scalers(X):
    """
    returns a list of sklearn.preprocessing.StandardScaler
    of input matrix.

    N, M = X.shape
    N = number of entries
    M = dimensionality
    """
    N, M = X.shape

    scalers = []

    # make a scaler for each dimension
    for i in range(M):
        scalers.append(StandardScaler())
        scalers[i].fit(X[:, i].reshape(-1, 1))

    return scalers


def apply_scaler(X, scalers):
    """
    applies the scalers to X
    N, M = X.shape
    N = number of entries
    M = dimensionality
    """
    N, M = X.shape

    X_scaled = np.zeros(shape=(N, M))

    for i in range(M):
        X_scaled[:, i] = scalers[i].transform(
            X[:, i].reshape(-1, 1)).reshape(1, -1)

    return X_scaled


def apply_inverse_scaler(X_scaled, scalers):
    """
    applies the inverse scalers to X_scaled
    N, M = X.shape
    N = number of entries
    M = dimensionality
    """
    N, M = X_scaled.shape

    X = np.zeros(shape=(N, M))

    for i in range(M):
        X[:, i] = scalers[i].inverse_transform(
            X_scaled[:, i].reshape(-1, 1)).reshape(1, -1)

    return X


def scale_data(X, y, validation_data=None):
    warnings.warn("function 'scale_data' is deprecated", DeprecationWarning)

    # domain
    N, M = X.shape
    # N = number of data points
    # M = dimension

    X_scalers = []
    X_scaled = np.zeros(shape=(N, M))

    for i in range(M):
        X_scalers.append(StandardScaler())
        X_scaled[:, i] = X_scalers[i].fit_transform(
            X[:, i].reshape(-1, 1)).reshape(-1,)

    # range
    # currently only have 1 y output
    y_scaler = StandardScaler()
    y_scaled = y_scaler.fit_transform(y.reshape(-1, 1)).reshape(-1,)

    if validation_data:
        Xval, yval = validation_data

        Nval, Mval = Xval.shape
        assert M == Mval, "train and validation M not the same"
        Xval_scaled = np.zeros(shape=(Nval, Mval))
        for i in range(Mval):
            Xval_scaled[:, i] = X_scalers[i].transform(
                Xval[:, i].reshape(-1, 1)).reshape(-1,)

        yval_scaled = y_scaler.transform(yval.reshape(-1, 1)).reshape(-1,)

        return X_scaled, y_scaled, Xval_scaled, yval_scaled
    else:
        return X_scaled, y_scaled



"""
module that contains helper functions to generate waveforms
"""

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as IUS
import phenom
import lal
import lalsimulation as lalsim
from scrinet.interfaces import lalutils, rotations
import functools
import schwimmbad

from scrinet.interfaces.rotations import spline_antidiff, spline_diff


class SinglePool(object):
    """
    from pycbc/pool.py
    used for when n_cores = 1
    """

    def broadcast(self, fcn, args):
        return self.map(fcn, [args])

    def map(self, f, items):
        return [f(a) for a in items]


def cart_to_polar(x, y, z):
    """
    cartesian to spherical polar transformation.
    phi (azimuthal angle) between [0, 2*pi]
    returns: r, theta, phi
    """
    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    theta = np.arctan2(hxy, z)
    phi = np.arctan2(y, x)
    phi = phi % (2 * np.pi)
    return r, theta, phi


def polar_to_cart(r, theta, phi):
    """
    spherical polar to cartesian transformation
    returns: x, y, z
    """
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)
    return x, y, z

# 1d non-spinning


def worker_wfgen_1d(task, new_times, M, deltaT, f_min, approximant, t_min, t_max, npts):
    q = task

    new_times = np.linspace(t_min, t_max, npts)

    m1, m2 = phenom.m1_m2_M_q(M, q)
    pp = dict(m1=m1, m2=m2, deltaT=deltaT,
              f_min=f_min,
              approximant=approximant
              )
    p = lalutils.gen_td_modes_wf_params(**pp)
    t, hlms = lalutils.gen_td_modes_wf(p, modes=[[2, 2]])

    # convert to dimensionless units
    t = phenom.StoM(t, m1+m2)
    # get 2,2 mode as we only support this.
    h22 = hlms[(2, 2)]

    if t_min < t[0]:
        print(
            f"possible interpolation issue: t_min ({t_min}) < t[0] ({t[0]:.2f})", flush=True)

    mask = (t >= t_min) & (t < t_max)
    t = t[mask]
    h22 = h22[mask]

    amp = np.abs(h22)
    phase = np.unwrap(np.angle(h22))

    iamp = IUS(t, amp)
    iphase = IUS(t, phase)

    amp = iamp(new_times)
    phase = iphase(new_times)
    freq = IUS(new_times, phase).derivative()(new_times)

    tshift = lalutils.peak_align_shift(new_times, amp)
    amp = lalutils.peak_align_interp(new_times, amp, tshift)
    phase = lalutils.peak_align_interp(new_times, phase, tshift)
    freq = lalutils.peak_align_interp(new_times, freq, tshift)

    return amp, phase, freq


def gen_1d_massratio_data(
        q_array,
        M,
        n_cores,
        deltaT=1/1024.,
        f_min=10,
        approximant=lalsim.SEOBNRv4P,
        t_min=-10000,
        t_max=100,
        npts=5000):

    new_times = np.linspace(t_min, t_max, npts)

    worker_wfgen_partial = functools.partial(worker_wfgen_1d,
                                             new_times=new_times,
                                             M=M,
                                             deltaT=deltaT,
                                             f_min=f_min,
                                             approximant=approximant,
                                             t_min=t_min,
                                             t_max=t_max,
                                             npts=npts)

    data = []
    coords = []

    tasks = list(q_array)

    if n_cores == 1:
        pool = SinglePool()
    else:
        pool = schwimmbad.choose_pool(mpi=False, processes=n_cores)
    results = pool.map(worker_wfgen_partial, tasks)
    try:
        pool.close
    except AttributeError:
        pass

    for i in range(len(q_array)):
        q = q_array[i]
        amp = results[i][0]
        phase = results[i][1]
        freq = results[i][2]

        d = {"t": new_times, "amp": amp, "phase": phase, "freq": freq}

        data.append(d)
        coords.append([q])

    n_t_points = len(new_times)
    n_waveforms = len(data)
    ts_amp = np.zeros(shape=(n_waveforms, n_t_points))
    ts_phase = np.zeros(shape=(n_waveforms, n_t_points))
    ts_freq = np.zeros(shape=(n_waveforms, n_t_points))
    ts_hreal = np.zeros(shape=(n_waveforms, n_t_points))
    ts_himag = np.zeros(shape=(n_waveforms, n_t_points))

    for i in range(n_waveforms):
        amp_pre_fac = phenom.eta_from_q(
            coords[i][0]) * lalutils.td_amp_scale(M, 1)
        ts_amp[i] = data[i]['amp'] / amp_pre_fac
        ts_phase[i] = data[i]['phase'] - data[i]['phase'][0]
        ts_freq[i] = data[i]['freq']

        h = ts_amp[i]*np.exp(-1.j * ts_phase[i])
        ts_hreal[i] = np.real(h)
        ts_himag[i] = np.imag(h)

    return new_times, ts_amp, ts_phase, ts_freq, ts_hreal, ts_himag, np.array(coords)

# 2d non-precessing, equal spin on both BHs


def worker_wfgen_2d(task, new_times, M, deltaT, f_min, approximant, t_min, t_max, npts):
    q, chi = task

    new_times = np.linspace(t_min, t_max, npts)

    m1, m2 = phenom.m1_m2_M_q(M, q)
    pp = dict(m1=m1, m2=m2, deltaT=deltaT,
              f_min=f_min,
              S1z=chi, S2z=chi,
              approximant=approximant
              )
    p = lalutils.gen_td_modes_wf_params(**pp)
    t, hlms = lalutils.gen_td_modes_wf(p, modes=[[2, 2]])

    # convert to dimensionless units
    t = phenom.StoM(t, m1+m2)
    # get 2,2 mode as we only support this.
    h22 = hlms[(2, 2)]

    if t_min < t[0]:
        print(
            f"possible interpolation issue: t_min ({t_min}) < t[0] ({t[0]:.2f})", flush=True)

    mask = (t >= t_min) & (t < t_max)
    t = t[mask]
    h22 = h22[mask]

    amp = np.abs(h22)
    phase = np.unwrap(np.angle(h22))

    iamp = IUS(t, amp)
    iphase = IUS(t, phase)

    amp = iamp(new_times)
    phase = iphase(new_times)
    freq = IUS(new_times, phase).derivative()(new_times)

    tshift = lalutils.peak_align_shift(new_times, amp)
    amp = lalutils.peak_align_interp(new_times, amp, tshift)
    phase = lalutils.peak_align_interp(new_times, phase, tshift)
    freq = lalutils.peak_align_interp(new_times, freq, tshift)

    return amp, phase, freq


def gen_2d_massratio_chi_data(
        q_array,
        chi_array,
        M,
        n_cores,
        deltaT=1/1024.,
        f_min=10,
        approximant=lalsim.SEOBNRv4P,
        t_min=-10000,
        t_max=100,
        npts=5000):

    new_times = np.linspace(t_min, t_max, npts)

    worker_wfgen_partial = functools.partial(worker_wfgen_2d,
                                             new_times=new_times,
                                             M=M,
                                             deltaT=deltaT,
                                             f_min=f_min,
                                             approximant=approximant,
                                             t_min=t_min,
                                             t_max=t_max,
                                             npts=npts)

    data = []
    coords = []

    tasks = list(zip(q_array, chi_array))

    if n_cores == 1:
        pool = SinglePool()
    else:
        pool = schwimmbad.choose_pool(mpi=False, processes=n_cores)
    results = pool.map(worker_wfgen_partial, tasks)
    try:
        pool.close
    except AttributeError:
        pass

    for i in range(len(q_array)):
        q = q_array[i]
        chi = chi_array[i]
        amp = results[i][0]
        phase = results[i][1]
        freq = results[i][2]

        d = {"t": new_times, "amp": amp, "phase": phase, "freq": freq}

        data.append(d)
        coords.append([q, chi])

    n_t_points = len(new_times)
    n_waveforms = len(data)
    ts_amp = np.zeros(shape=(n_waveforms, n_t_points))
    ts_phase = np.zeros(shape=(n_waveforms, n_t_points))
    ts_freq = np.zeros(shape=(n_waveforms, n_t_points))
    ts_hreal = np.zeros(shape=(n_waveforms, n_t_points))
    ts_himag = np.zeros(shape=(n_waveforms, n_t_points))

    for i in range(n_waveforms):
        amp_pre_fac = phenom.eta_from_q(
            coords[i][0]) * lalutils.td_amp_scale(M, 1)
        ts_amp[i] = data[i]['amp'] / amp_pre_fac
        ts_phase[i] = data[i]['phase'] - data[i]['phase'][0]
        ts_freq[i] = data[i]['freq']
        h = ts_amp[i]*np.exp(-1.j * ts_phase[i])
        ts_hreal[i] = np.real(h)
        ts_himag[i] = np.imag(h)

    return new_times, ts_amp, ts_phase, ts_freq, ts_hreal, ts_himag, np.array(coords)


# 3d non-precessing

def worker_wfgen_3d_non_prec(task, new_times, M, deltaT, f_min, approximant, t_min, t_max, npts):
    q, chi1z, chi2z = task

    new_times = np.linspace(t_min, t_max, npts)

    m1, m2 = phenom.m1_m2_M_q(M, q)
    pp = dict(m1=m1, m2=m2, deltaT=deltaT,
              f_min=f_min,
              S1z=chi1z, S2z=chi2z,
              approximant=approximant
              )
    p = lalutils.gen_td_modes_wf_params(**pp)
    t, hlms = lalutils.gen_td_modes_wf(p, modes=[[2, 2]])

    # convert to dimensionless units
    t = phenom.StoM(t, m1+m2)
    # get 2,2 mode as we only support this.
    h22 = hlms[(2, 2)]

    if t_min < t[0]:
        print(
            f"possible interpolation issue: t_min ({t_min}) < t[0] ({t[0]:.2f})", flush=True)

    mask = (t >= t_min) & (t < t_max)
    t = t[mask]
    h22 = h22[mask]

    amp = np.abs(h22)
    phase = np.unwrap(np.angle(h22))

    iamp = IUS(t, amp)
    iphase = IUS(t, phase)

    amp = iamp(new_times)
    phase = iphase(new_times)
    freq = IUS(new_times, phase).derivative()(new_times)

    tshift = lalutils.peak_align_shift(new_times, amp)
    amp = lalutils.peak_align_interp(new_times, amp, tshift)
    phase = lalutils.peak_align_interp(new_times, phase, tshift)
    freq = lalutils.peak_align_interp(new_times, freq, tshift)

    return amp, phase, freq


def gen_3d_non_prec_data(
        q_array,
        chi1z_array,
        chi2z_array,
        M,
        n_cores,
        deltaT=1/1024.,
        f_min=10,
        approximant=lalsim.SEOBNRv4P,
        t_min=-10000,
        t_max=100,
        npts=5000):

    new_times = np.linspace(t_min, t_max, npts)

    worker_wfgen_partial = functools.partial(worker_wfgen_3d_non_prec,
                                             new_times=new_times,
                                             M=M,
                                             deltaT=deltaT,
                                             f_min=f_min,
                                             approximant=approximant,
                                             t_min=t_min,
                                             t_max=t_max,
                                             npts=npts)

    data = []
    coords = []

    tasks = list(zip(q_array, chi1z_array, chi2z_array))

    if n_cores == 1:
        pool = SinglePool()
    else:
        pool = schwimmbad.choose_pool(mpi=False, processes=n_cores)
    results = pool.map(worker_wfgen_partial, tasks)
    try:
        pool.close
    except AttributeError:
        pass

    for i in range(len(q_array)):
        q = q_array[i]
        chi1z = chi1z_array[i]
        chi2z = chi2z_array[i]
        amp = results[i][0]
        phase = results[i][1]
        freq = results[i][2]

        d = {"t": new_times, "amp": amp, "phase": phase, "freq": freq}

        data.append(d)
        coords.append([q, chi1z, chi2z])

    n_t_points = len(new_times)
    n_waveforms = len(data)
    ts_amp = np.zeros(shape=(n_waveforms, n_t_points))
    ts_phase = np.zeros(shape=(n_waveforms, n_t_points))
    ts_freq = np.zeros(shape=(n_waveforms, n_t_points))
    ts_hreal = np.zeros(shape=(n_waveforms, n_t_points))
    ts_himag = np.zeros(shape=(n_waveforms, n_t_points))

    for i in range(n_waveforms):
        amp_pre_fac = phenom.eta_from_q(
            coords[i][0]) * lalutils.td_amp_scale(M, 1)
        ts_amp[i] = data[i]['amp'] / amp_pre_fac
        ts_phase[i] = data[i]['phase'] - data[i]['phase'][0]
        ts_freq[i] = data[i]['freq']
        h = ts_amp[i]*np.exp(-1.j * ts_phase[i])
        ts_hreal[i] = np.real(h)
        ts_himag[i] = np.imag(h)

    return new_times, ts_amp, ts_phase, ts_freq, ts_hreal, ts_himag, np.array(coords)

# 3d precession single spin


def worker_wfgen_3d_prec_single_spin(task, new_times, M, deltaT, f_min, approximant, t_min, t_max, npts, modes):
    q, chi1x, chi1z = task

    new_times = np.linspace(t_min, t_max, npts)

    m1, m2 = phenom.m1_m2_M_q(M, q)
    pp = dict(m1=m1, m2=m2, deltaT=deltaT,
              f_min=f_min,
              f_ref=f_min,
              S1x=chi1x, S1z=chi1z,
              approximant=approximant
              )
    p = lalutils.gen_td_modes_wf_params(**pp)

    t, hlms = lalutils.gen_td_modes_wf(p, modes=modes)

    # convert to dimensionless units
    t = phenom.StoM(t, m1+m2)

    # compute peak time from 2,2 mode for now - this is a simplification
    # that should be changed later

    h22 = hlms[(2, 2)]

    if t_min < t[0]:
        print(
            f"possible interpolation issue: t_min ({t_min}) < t[0] ({t[0]:.2f})", flush=True)

    mask = (t >= t_min) & (t < t_max)
    t = t[mask]
    h22 = h22[mask]

    amp = np.abs(h22)
    phase = np.unwrap(np.angle(h22))

    iamp = IUS(t, amp)
    iphase = IUS(t, phase)

    amp = iamp(new_times)
    phase = iphase(new_times)
    freq = IUS(new_times, phase).derivative()(new_times)

    tshift = lalutils.peak_align_shift(new_times, amp)
    amp = lalutils.peak_align_interp(new_times, amp, tshift)
    phase = lalutils.peak_align_interp(new_times, phase, tshift)
    freq = lalutils.peak_align_interp(new_times, freq, tshift)

    return amp, phase, freq


def gen_3d_prec_single_spin_data(
    q_array,
    chi1_array,
    theta1_array,
    M,
    n_cores,
    deltaT=1/1024.,
    f_min=10,
    approximant=lalsim.SEOBNRv4P,
    t_min=-10000,
    t_max=100,
    npts=5000,
    modes=[[2, 2]]
):
    """
    parameters:
        mass-ratio
        spin1 magnitude
        theta1 (polar angle of spin)
    """
    # convert polar to cartesian
    phi1_array = np.zeros(len(chi1_array))
    chi1x_array, chi1y_array, chi1z_array = polar_to_cart(
        chi1_array, theta1_array, phi1_array)

    new_times = np.linspace(t_min, t_max, npts)

    # just modelling 2,2 interial frame mode for now to see what it looks like

    worker_wfgen_partial = functools.partial(worker_wfgen_3d_prec_single_spin,
                                             new_times=new_times,
                                             M=M,
                                             deltaT=deltaT,
                                             f_min=f_min,
                                             approximant=approximant,
                                             t_min=t_min,
                                             t_max=t_max,
                                             npts=npts,
                                             modes=modes)

    data = []
    coords = []

    tasks = list(zip(q_array, chi1x_array, chi1z_array))

    if n_cores == 1:
        pool = SinglePool()
    else:
        pool = schwimmbad.choose_pool(mpi=False, processes=n_cores)
    results = pool.map(worker_wfgen_partial, tasks)
    try:
        pool.close
    except AttributeError:
        pass

    for i in range(len(q_array)):
        q = q_array[i]
        chi1x = chi1x_array[i]
        chi1z = chi1z_array[i]
        amp = results[i][0]
        phase = results[i][1]
        freq = results[i][2]

        d = {"t": new_times, "amp": amp, "phase": phase, "freq": freq}

        data.append(d)
        coords.append([q, chi1x, chi1z])

    n_t_points = len(new_times)
    n_waveforms = len(data)
    ts_amp = np.zeros(shape=(n_waveforms, n_t_points))
    ts_phase = np.zeros(shape=(n_waveforms, n_t_points))
    ts_freq = np.zeros(shape=(n_waveforms, n_t_points))
    ts_hreal = np.zeros(shape=(n_waveforms, n_t_points))
    ts_himag = np.zeros(shape=(n_waveforms, n_t_points))

    for i in range(n_waveforms):
        amp_pre_fac = phenom.eta_from_q(
            coords[i][0]) * lalutils.td_amp_scale(M, 1)
        ts_amp[i] = data[i]['amp'] / amp_pre_fac
        ts_phase[i] = data[i]['phase']  # - data[i]['phase'][0]
        ts_freq[i] = data[i]['freq']
        h = ts_amp[i]*np.exp(-1.j * ts_phase[i])
        ts_hreal[i] = np.real(h)
        ts_himag[i] = np.imag(h)

    return new_times, ts_amp, ts_phase, ts_freq, ts_hreal, ts_himag, np.array(coords)

# 3d precession single spin - coprecessing frame


def worker_wfgen_3d_prec_single_spin_coprec(task, new_times, M, deltaT, f_min, approximant, t_min, t_max, npts, modes):
    q, chi1x, chi1z = task

    new_times = np.linspace(t_min, t_max, npts)

    m1, m2 = phenom.m1_m2_M_q(M, q)
    pp = dict(m1=m1, m2=m2, deltaT=deltaT,
              f_min=f_min,
              S1x=chi1x, S1z=chi1z,
              approximant=approximant
              )
    p = lalutils.gen_td_modes_wf_params(**pp)

    if approximant in [lalsim.SEOBNRv4PHM]:
        t, hlms = lalutils.gen_td_modes_wf(
            p, eob_all_ell_2_modes=True, modes=[[2, 2], [2, 1]])
    else:
        t, hlms = lalutils.gen_td_modes_wf(p, modes=modes)

    # convert to dimensionless units
    t = phenom.StoM(t, m1+m2)

    # estimate thetaJN and rotate to J-aligned frame
    thetaJN, alpha0, phi_aligned = rotations.compute_L_to_J_angles(
        p['m1'],
        p['m2'],
        p['f_ref'],
        0,
        p['phiRef'],
        p['S1x'],
        p['S1y'],
        p['S1z'],
        p['S2x'],
        p['S2y'],
        p['S2z'])

    wr = rotations.WaveformRotations(
        times=t, hlms=hlms, frame='inertial-L', alpha0=alpha0, thetaJN=thetaJN, phi0=phi_aligned)
    wr.from_inertial_frame_to_coprecessing_frame()
    hlms_coprec = wr.hlms
    alpha = wr.alpha
    beta = wr.beta
    gamma = wr.gamma

    # compute peak time from 2,2 mode for now - this is a simplification
    # that should be changed later

    h22 = hlms_coprec[(2, 2)]

    if t_min < t[0]:
        print(
            f"possible interpolation issue: t_min ({t_min}) < t[0] ({t[0]:.2f})", flush=True)

    mask = (t >= t_min) & (t < t_max)
    t = t[mask]
    h22 = h22[mask]
    alpha = alpha[mask]
    beta = beta[mask]
    # gamma = gamma[mask]

    amp = np.abs(h22)
    phase = np.unwrap(np.angle(h22))

    iamp = IUS(t, amp)
    iphase = IUS(t, phase)

    amp = iamp(new_times)
    phase = iphase(new_times)
    freq = IUS(new_times, phase).derivative()(new_times)

    alpha = IUS(t, alpha)(new_times)
    beta = IUS(t, beta)(new_times)
    # gamma = IUS(t, gamma)(new_times)

    tshift = lalutils.peak_align_shift(new_times, amp)
    amp = lalutils.peak_align_interp(new_times, amp, tshift)
    phase = lalutils.peak_align_interp(new_times, phase, tshift)
    freq = lalutils.peak_align_interp(new_times, freq, tshift)

    alpha = lalutils.peak_align_interp(new_times, alpha, tshift)

    # shift alpha to try and make the alphas have some self similarity
    alpha -= alpha[0]

    beta = lalutils.peak_align_interp(new_times, beta, tshift)
    # gamma = lalutils.peak_align_interp(new_times, gamma, tshift)

    # do this so that gamma is calculated from the alpha that has been shifted
    k = 1
    gamma = - spline_antidiff(new_times, np.cos(beta)
                              * spline_diff(new_times, alpha, k=k), k=k)
    gamma = np.unwrap(gamma)

    # beta -= thetaJN

    return amp, phase, freq, alpha, beta, gamma


def gen_3d_prec_single_spin_coprec_data(
        q_array,
        chi1_array,
        theta1_array,
        M,
        n_cores,
        deltaT=1/1024.,
        f_min=10,
        approximant=lalsim.SEOBNRv4PHM,
        t_min=-10000,
        t_max=100,
        npts=5000):
    """
    parameters:
        mass-ratio
        spin1 magnitude
        theta1 (polar angle of spin)
    """
    # convert polar to cartesian
    phi1_array = np.zeros(len(chi1_array))
    chi1x_array, chi1y_array, chi1z_array = polar_to_cart(
        chi1_array, theta1_array, phi1_array)

    new_times = np.linspace(t_min, t_max, npts)

    # just modelling 2,2 interial frame mode for now to see what it looks like
    # default_modes = [[2,2], [2,1], [2,0], [2,-1], [2,-2]]
    default_modes = [[2, 2]]

    worker_wfgen_partial = functools.partial(worker_wfgen_3d_prec_single_spin_coprec,
                                             new_times=new_times,
                                             M=M,
                                             deltaT=deltaT,
                                             f_min=f_min,
                                             approximant=approximant,
                                             t_min=t_min,
                                             t_max=t_max,
                                             npts=npts,
                                             modes=default_modes)

    data = []
    coords = []

    tasks = list(zip(q_array, chi1x_array, chi1z_array))

    if n_cores == 1:
        pool = SinglePool()
    else:
        pool = schwimmbad.choose_pool(mpi=False, processes=n_cores)
    results = pool.map(worker_wfgen_partial, tasks)
    try:
        pool.close
    except AttributeError:
        pass

    for i in range(len(q_array)):
        q = q_array[i]
        chi1x = chi1x_array[i]
        chi1z = chi1z_array[i]
        amp = results[i][0]
        phase = results[i][1]
        freq = results[i][2]
        alpha = results[i][3]
        beta = results[i][4]
        gamma = results[i][5]

        d = {"t": new_times, "amp": amp, "phase": phase, "freq": freq,
             "alpha": alpha, "beta": beta, "gamma": gamma
             }

        data.append(d)
        coords.append([q, chi1x, chi1z])

    n_t_points = len(new_times)
    n_waveforms = len(data)
    ts_amp = np.zeros(shape=(n_waveforms, n_t_points))
    ts_phase = np.zeros(shape=(n_waveforms, n_t_points))
    ts_freq = np.zeros(shape=(n_waveforms, n_t_points))
    ts_hreal = np.zeros(shape=(n_waveforms, n_t_points))
    ts_himag = np.zeros(shape=(n_waveforms, n_t_points))

    ts_alpha = np.zeros(shape=(n_waveforms, n_t_points))
    ts_beta = np.zeros(shape=(n_waveforms, n_t_points))
    ts_gamma = np.zeros(shape=(n_waveforms, n_t_points))

    for i in range(n_waveforms):
        amp_pre_fac = phenom.eta_from_q(
            coords[i][0]) * lalutils.td_amp_scale(M, 1)
        ts_amp[i] = data[i]['amp'] / amp_pre_fac
        ts_phase[i] = data[i]['phase'] - data[i]['phase'][0]
        ts_freq[i] = data[i]['freq']
        h = ts_amp[i]*np.exp(-1.j * ts_phase[i])
        ts_hreal[i] = np.real(h)
        ts_himag[i] = np.imag(h)

        ts_alpha[i] = data[i]['alpha']
        ts_beta[i] = data[i]['beta']
        ts_gamma[i] = data[i]['gamma']

    return new_times, ts_amp, ts_phase, ts_freq, ts_hreal, ts_himag, ts_alpha, ts_beta, ts_gamma, np.array(coords)


# 7d precession single spin - coprecessing frame

def worker_wfgen_7d_prec_single_spin_coprec(task, new_times, M, deltaT, f_min, approximant, t_min, t_max, npts, modes):
    q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z = task

    new_times = np.linspace(t_min, t_max, npts)

    m1, m2 = phenom.m1_m2_M_q(M, q)
    pp = dict(m1=m1, m2=m2, deltaT=deltaT,
              f_min=f_min,
              S1x=chi1x, S1y=chi1y, S1z=chi1z,
              S2x=chi2x, S2y=chi2y, S2z=chi2z,
              approximant=approximant
              )
    p = lalutils.gen_td_modes_wf_params(**pp)

    if approximant in [lalsim.SEOBNRv4PHM]:
        t, hlms = lalutils.gen_td_modes_wf(
            p, eob_all_ell_2_modes=True, modes=[[2, 2], [2, 1]])
    else:
        t, hlms = lalutils.gen_td_modes_wf(p, modes=modes)

    # convert to dimensionless units
    t = phenom.StoM(t, m1+m2)

    # estimate thetaJN and rotate to J-aligned frame
    thetaJN, alpha0, phi_aligned = rotations.compute_L_to_J_angles(
        p['m1'],
        p['m2'],
        p['f_ref'],
        0,
        p['phiRef'],
        p['S1x'],
        p['S1y'],
        p['S1z'],
        p['S2x'],
        p['S2y'],
        p['S2z'])

    wr = rotations.WaveformRotations(
        times=t, hlms=hlms, frame='inertial-L', alpha0=alpha0, thetaJN=thetaJN, phi0=phi_aligned)
    wr.from_inertial_frame_to_coprecessing_frame()
    hlms_coprec = wr.hlms
    alpha = wr.alpha
    beta = wr.beta
    gamma = wr.gamma

    # compute peak time from 2,2 mode for now - this is a simplification
    # that should be changed later

    h22 = hlms_coprec[(2, 2)]

    if t_min < t[0]:
        print(
            f"possible interpolation issue: t_min ({t_min}) < t[0] ({t[0]:.2f})", flush=True)

    mask = (t >= t_min) & (t < t_max)
    t = t[mask]
    h22 = h22[mask]
    alpha = alpha[mask]
    beta = beta[mask]
    # gamma = gamma[mask]

    amp = np.abs(h22)
    phase = np.unwrap(np.angle(h22))

    iamp = IUS(t, amp)
    iphase = IUS(t, phase)

    amp = iamp(new_times)
    phase = iphase(new_times)
    freq = IUS(new_times, phase).derivative()(new_times)

    alpha = IUS(t, alpha)(new_times)
    beta = IUS(t, beta)(new_times)
    # gamma = IUS(t, gamma)(new_times)

    tshift = lalutils.peak_align_shift(new_times, amp)
    amp = lalutils.peak_align_interp(new_times, amp, tshift)
    phase = lalutils.peak_align_interp(new_times, phase, tshift)
    freq = lalutils.peak_align_interp(new_times, freq, tshift)

    alpha = lalutils.peak_align_interp(new_times, alpha, tshift)

    # shift alpha to try and make the alphas have some self similarity
    alpha -= alpha[0]

    beta = lalutils.peak_align_interp(new_times, beta, tshift)
    # gamma = lalutils.peak_align_interp(new_times, gamma, tshift)

    # do this so that gamma is calculated from the alpha that has been shifted
    k = 1
    gamma = - spline_antidiff(new_times, np.cos(beta)
                              * spline_diff(new_times, alpha, k=k), k=k)
    gamma = np.unwrap(gamma)

    # beta -= thetaJN

    return amp, phase, freq, alpha, beta, gamma


def gen_7d_prec_single_spin_coprec_data(
        q_array,
        chi1_array,
        theta1_array,
        phi1_array,
        chi2_array,
        theta2_array,
        phi2_array,
        M,
        n_cores,
        deltaT=1/1024.,
        f_min=10,
        approximant=lalsim.SEOBNRv4PHM,
        t_min=-10000,
        t_max=100,
        npts=5000):
    """
    parameters:
        mass-ratio
        spin1 magnitude
        theta1 (polar angle of spin 1)
        phi1 (azimuthal angle of spin 1)
        spin2 magnitude
        theta2 (polar angle of spin 2)
        phi2 (azimuthal angle of spin 2)
    """
    # convert polar to cartesian
    chi1x_array, chi1y_array, chi1z_array = polar_to_cart(
        chi1_array, theta1_array, phi1_array)

    chi2x_array, chi2y_array, chi2z_array = polar_to_cart(
        chi2_array, theta2_array, phi2_array)

    new_times = np.linspace(t_min, t_max, npts)

    # just modelling 2,2 interial frame mode for now to see what it looks like
    # default_modes = [[2,2], [2,1], [2,0], [2,-1], [2,-2]]
    default_modes = [[2, 2]]

    worker_wfgen_partial = functools.partial(worker_wfgen_7d_prec_single_spin_coprec,
                                             new_times=new_times,
                                             M=M,
                                             deltaT=deltaT,
                                             f_min=f_min,
                                             approximant=approximant,
                                             t_min=t_min,
                                             t_max=t_max,
                                             npts=npts,
                                             modes=default_modes)

    data = []
    coords = []

    tasks = list(zip(q_array, chi1x_array, chi1y_array,
                     chi1z_array, chi2x_array, chi2y_array, chi2z_array))

    if n_cores == 1:
        pool = SinglePool()
    else:
        pool = schwimmbad.choose_pool(mpi=False, processes=n_cores)
    results = pool.map(worker_wfgen_partial, tasks)
    try:
        pool.close
    except AttributeError:
        pass

    for i in range(len(q_array)):
        q = q_array[i]
        chi1x = chi1x_array[i]
        chi1y = chi1y_array[i]
        chi1z = chi1z_array[i]
        chi2x = chi2x_array[i]
        chi2y = chi2y_array[i]
        chi2z = chi2z_array[i]
        amp = results[i][0]
        phase = results[i][1]
        freq = results[i][2]
        alpha = results[i][3]
        beta = results[i][4]
        gamma = results[i][5]

        d = {"t": new_times, "amp": amp, "phase": phase, "freq": freq,
             "alpha": alpha, "beta": beta, "gamma": gamma
             }

        data.append(d)
        coords.append([q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z])

    n_t_points = len(new_times)
    n_waveforms = len(data)
    ts_amp = np.zeros(shape=(n_waveforms, n_t_points))
    ts_phase = np.zeros(shape=(n_waveforms, n_t_points))
    ts_freq = np.zeros(shape=(n_waveforms, n_t_points))
    ts_hreal = np.zeros(shape=(n_waveforms, n_t_points))
    ts_himag = np.zeros(shape=(n_waveforms, n_t_points))

    ts_alpha = np.zeros(shape=(n_waveforms, n_t_points))
    ts_beta = np.zeros(shape=(n_waveforms, n_t_points))
    ts_gamma = np.zeros(shape=(n_waveforms, n_t_points))

    for i in range(n_waveforms):
        amp_pre_fac = phenom.eta_from_q(
            coords[i][0]) * lalutils.td_amp_scale(M, 1)
        ts_amp[i] = data[i]['amp'] / amp_pre_fac
        ts_phase[i] = data[i]['phase'] - data[i]['phase'][0]
        ts_freq[i] = data[i]['freq']
        h = ts_amp[i]*np.exp(-1.j * ts_phase[i])
        ts_hreal[i] = np.real(h)
        ts_himag[i] = np.imag(h)

        ts_alpha[i] = data[i]['alpha']
        ts_beta[i] = data[i]['beta']
        ts_gamma[i] = data[i]['gamma']

    return new_times, ts_amp, ts_phase, ts_freq, ts_hreal, ts_himag, ts_alpha, ts_beta, ts_gamma, np.array(coords)



import itertools
import sys
import numpy as np
import scipy.stats
from matplotlib import (offsetbox, pyplot, gridspec)


def create_axes_grid(parameters, labels=None, height_ratios=None,
                     width_ratios=None, no_diagonals=False):
    """
    Taken from gwastro/pycbc
    https://github.com/gwastro/pycbc/blob/master/pycbc/results/scatter_histograms.py

    Given a list of parameters, creates a figure with an axis for
    every possible combination of the parameters.
    Parameters
    ----------
    parameters : list
        Names of the variables to be plotted.
    labels : {None, dict}, optional
        A dictionary of parameters -> parameter labels.
    height_ratios : {None, list}, optional
        Set the height ratios of the axes; see `matplotlib.gridspec.GridSpec`
        for details.
    width_ratios : {None, list}, optional
        Set the width ratios of the axes; see `matplotlib.gridspec.GridSpec`
        for details.
    no_diagonals : {False, bool}, optional
        Do not produce axes for the same parameter on both axes.
    Returns
    -------
    fig : pyplot.figure
        The figure that was created.
    axis_dict : dict
        A dictionary mapping the parameter combinations to the axis and their
        location in the subplots grid; i.e., the key, values are:
        `{('param1', 'param2'): (pyplot.axes, row index, column index)}`
    """
    if labels is None:
        labels = {p: p for p in parameters}
    elif any(p not in labels for p in parameters):
        raise ValueError("labels must be provided for all parameters")
    # Create figure with adequate size for number of parameters.
    ndim = len(parameters)
    if no_diagonals:
        ndim -= 1
    if ndim < 3:
        fsize = (8, 7)
    else:
        fsize = (ndim*3 - 1, ndim*3 - 2)
    fig = pyplot.figure(figsize=fsize)
    # create the axis grid
    gs = gridspec.GridSpec(ndim, ndim, width_ratios=width_ratios,
                           height_ratios=height_ratios,
                           wspace=0.05, hspace=0.05)
    # create grid of axis numbers to easily create axes in the right locations
    axes = np.arange(ndim**2).reshape((ndim, ndim))

    # Select possible combinations of plots and establish rows and columns.
    combos = list(itertools.combinations(parameters, 2))
    # add the diagonals
    if not no_diagonals:
        combos += [(p, p) for p in parameters]

    # create the mapping between parameter combos and axes
    axis_dict = {}
    # cycle over all the axes, setting thing as needed
    for nrow in range(ndim):
        for ncolumn in range(ndim):
            ax = pyplot.subplot(gs[axes[nrow, ncolumn]])
            # map to a parameter index
            px = parameters[ncolumn]
            if no_diagonals:
                py = parameters[nrow+1]
            else:
                py = parameters[nrow]
            if (px, py) in combos:
                axis_dict[px, py] = (ax, nrow, ncolumn)
                # x labels only on bottom
                if nrow + 1 == ndim:
                    ax.set_xlabel('{}'.format(labels[px]), fontsize=18)
                else:
                    pyplot.setp(ax.get_xticklabels(), visible=False)
                # y labels only on left
                if ncolumn == 0:
                    ax.set_ylabel('{}'.format(labels[py]), fontsize=18)
                else:
                    pyplot.setp(ax.get_yticklabels(), visible=False)
            else:
                # make non-used axes invisible
                ax.axis('off')
    return fig, axis_dict


def get_scale_fac(fig, fiducial_width=8, fiducial_height=7):
    """
    Taken from gwastro/pycbc
    https://github.com/gwastro/pycbc/blob/master/pycbc/results/scatter_histograms.py

    Gets a factor to scale fonts by for the given figure. The scale
    factor is relative to a figure with dimensions
    (`fiducial_width`, `fiducial_height`).
    """
    width, height = fig.get_size_inches()
    return (width*height/(fiducial_width*fiducial_height))**0.5


def reduce_ticks(ax, which, maxticks=3):
    """
    Taken from gwastro/pycbc
    https://github.com/gwastro/pycbc/blob/master/pycbc/results/scatter_histograms.py

    Given a pyplot axis, resamples its `which`-axis ticks such that are at most
    `maxticks` left.
    Parameters
    ----------
    ax : axis
        The axis to adjust.
    which : {'x' | 'y'}
        Which axis to adjust.
    maxticks : {3, int}
        Maximum number of ticks to use.
    Returns
    -------
    array
        An array of the selected ticks.
    """
    ticks = getattr(ax, 'get_{}ticks'.format(which))()
    if len(ticks) > maxticks:
        # make sure the left/right value is not at the edge
        minax, maxax = getattr(ax, 'get_{}lim'.format(which))()
        dw = abs(maxax-minax)/10.
        start_idx, end_idx = 0, len(ticks)
        if ticks[0] < minax + dw:
            start_idx += 1
        if ticks[-1] > maxax - dw:
            end_idx -= 1
        # get reduction factor
        fac = int(len(ticks) / maxticks)
        ticks = ticks[start_idx:end_idx:fac]
    return ticks


def create_multidim_plot(parameters, samples, labels=None,
                         mins=None, maxs=None, expected_parameters=None,
                         expected_parameters_color='r',
                         plot_scatter=True,
                         zvals=None, show_colorbar=True, cbar_label=None,
                         vmin=None, vmax=None, scatter_cmap='viridis',
                         fig=None, axis_dict=None, cb_scale=10,
                         scatter_colour='k', scatter_size=5, scatter_marker='o'):
    """
    Taken from gwastro/pycbc
    https://github.com/gwastro/pycbc/blob/master/pycbc/results/scatter_histograms.py

    Generate a figure with several plots and histograms.
    Parameters
    ----------
    parameters: list
        Names of the variables to be plotted.
    samples : np Record Array
        A record array of the samples to plot.
    labels: dict, optional
        A dictionary mapping parameters to labels. If none provided, will just
        use the parameter strings as the labels.
    mins : {None, dict}, optional
        Minimum value for the axis of each variable in `parameters`.
        If None, it will use the minimum of the corresponding variable in
        `samples`.
    maxs : {None, dict}, optional
        Maximum value for the axis of each variable in `parameters`.
        If None, it will use the maximum of the corresponding variable in
        `samples`.
    expected_parameters : {None, dict}, optional
        Expected values of `parameters`, as a dictionary mapping parameter
        names -> values. A cross will be plotted at the location of the
        expected parameters on axes that plot any of the expected parameters.
    expected_parameters_color : {'r', string}, optional
        What color to make the expected parameters cross.
    plot_scatter : {True, bool}
        Plot each sample point as a scatter plot.
    zvals : {None, array}
        An array to use for coloring the scatter plots. If None, scatter points
        will be the same color.
    show_colorbar : {True, bool}
        Show the colorbar of zvalues used for the scatter points. A ValueError
        will be raised if zvals is None and this is True.
    cbar_label : {None, str}
        Specify a label to add to the colorbar.
    vmin: {None, float}, optional
        Minimum value for the colorbar. If None, will use the minimum of zvals.
    vmax: {None, float}, optional
        Maximum value for the colorbar. If None, will use the maxmimum of
        zvals.
    scatter_cmap : {'viridis', string}
        The color map to use for the scatter points. Default is 'viridis'.
    cb_scale : {10, int}
        scale tick size on colour bar
    scatter_colour : {'k', str}
        colour of points if no zvals data given
    scatter_size : {5, int}
        size of points
    scatter_marker : {'o', str}
        marker style for scatter plot
    Returns
    -------
    fig : pyplot.figure
        The figure that was created.
    axis_dict : dict
        A dictionary mapping the parameter combinations to the axis and their
        location in the subplots grid; i.e., the key, values are:
        `{('param1', 'param2'): (pyplot.axes, row index, column index)}`
    """
    if labels is None:
        labels = {p: p for p in parameters}
    # set up the figure with a grid of axes
    # if only plotting 2 parameters, make the marginal plots smaller
    nparams = len(parameters)
    if nparams == 2:
        width_ratios = [3, 1]
        height_ratios = [1, 3]
    else:
        width_ratios = height_ratios = None

    # only plot scatter if more than one parameter
    plot_scatter = plot_scatter and nparams > 1

    # Sort zvals to get higher values on top in scatter plots
    if plot_scatter:
        if zvals is not None:
            sort_indices = zvals.argsort()
            zvals = zvals[sort_indices]
            samples = samples[sort_indices]
        elif show_colorbar:
            raise ValueError("must provide z values to create a colorbar")
        else:
            # just make all scatter points same color
            zvals = scatter_colour

    # convert samples to a dictionary to avoid re-computing derived parameters
    # every time they are needed
    samples = dict([[p, samples[p]] for p in parameters])

    # values for axis bounds
    if mins is None:
        mins = {p: samples[p].min() for p in parameters}
    else:
        # copy the dict
        mins = {p: val for p, val in mins.items()}
    if maxs is None:
        maxs = {p: samples[p].max() for p in parameters}
    else:
        # copy the dict
        maxs = {p: val for p, val in maxs.items()}

    # create the axis grid
    if fig is None and axis_dict is None:
        fig, axis_dict = create_axes_grid(
            parameters, labels=labels,
            width_ratios=width_ratios, height_ratios=height_ratios,
            no_diagonals=True)

    # Off-diagonals...
    for px, py in axis_dict:
        if px == py:
            continue
        ax, _, _ = axis_dict[px, py]
        if plot_scatter:
            alpha = 1.
            plt = ax.scatter(x=samples[px], y=samples[py], c=zvals,
                             s=scatter_size, marker=scatter_marker,
                             edgecolors='none', vmin=vmin, vmax=vmax,
                             cmap=scatter_cmap, alpha=alpha, zorder=2)
        if expected_parameters is not None:
            try:
                ax.axvline(expected_parameters[px], lw=1.5,
                           color=expected_parameters_color, zorder=5)
            except KeyError:
                pass
            try:
                ax.axhline(expected_parameters[py], lw=1.5,
                           color=expected_parameters_color, zorder=5)
            except KeyError:
                pass

        ax.set_xlim(mins[px], maxs[px])
        ax.set_ylim(mins[py], maxs[py])

    # adjust tick number for large number of plots
    if len(parameters) > 3:
        for px, py in axis_dict:
            ax, _, _ = axis_dict[px, py]
            ax.set_xticks(reduce_ticks(ax, 'x', maxticks=3))
            ax.set_yticks(reduce_ticks(ax, 'y', maxticks=3))

    if plot_scatter and show_colorbar:
        # compute font size based on fig size
        scale_fac = get_scale_fac(fig)
        fig.subplots_adjust(right=0.85, wspace=0.03)
        cbar_ax = fig.add_axes([0.9, 0.1, 0.03, 0.8])
        cb = fig.colorbar(plt, cax=cbar_ax)
        if cbar_label is not None:
            cb.set_label(cbar_label, fontsize=12*scale_fac)
        cb.ax.tick_params(labelsize=cb_scale*scale_fac)

    return fig, axis_dict



import numpy as np

# Newton gravity constant
# lal.G_SI: 6.67384e-11
G_Newt = 6.67384e-11

# light speed
# lal.C_SI: 299792458.0
c_ls = 299792458.0

# lal.MSUN_SI: 1.9885469549614615e+30
MSUN_SI = 1.9885469549614615e+30

MTSUN_SI = 4.925491025543576e-06

# lal.MRSUN_SI: 1476.6250614046494
MRSUN_SI = 1476.6250614046494

# lal.PC_SI: 3.085677581491367e+16
PC_SI = 3.085677581491367e+16

# lal.GAMMA: 0.5772156649015329
GAMMA = 0.5772156649015329


def Msun_to_sec(M):
    """
    convert mass (in units of solar masses)
    into seconds
    """
#     return M *lal.MSUN_SI* G_Newt / c_ls**3.
    return M * MTSUN_SI


def TaylorT3_phase(t, tc, eta, M, phi0=0, m=2):
    """
    orbital phase, but returns by default the m=2 mode
    equation 3.10a in https://arxiv.org/pdf/0907.0700.pdf

    https://git.ligo.org/lscsoft/lalsuite/blob/master/lalsimulation/src/LALSimInspiralTaylorT3.c

    https://git.ligo.org/lscsoft/lalsuite/blob/master/lalsimulation/src/LALSimInspiralPNCoefficients.c

    t: time
    tc: coalescence time
    eta: symmetric mass ratio
    M: total mass (Msun)
    phi0: reference orbital phase, default 0
    m: m-mode default = 2
    """

    Msec = Msun_to_sec(M)
    Msec = M

    pi2 = np.pi*np.pi

    c1 = eta/(5.*Msec)

    td = c1 * (tc - t)

#     td = np.sqrt(td**2 + 1)

    theta = td**(-1./8.)  # -1./8. = -0.125

    theta2 = theta*theta
    theta3 = theta2*theta
    theta4 = theta3*theta
    theta5 = theta4*theta
    theta6 = theta5*theta
    theta7 = theta6*theta

    # pre factor
    ftaN = -1. / eta
    # 0PN
    fts1 = 1.
    # 0.5PN = 0 in GR
    # 1PN
    fta2 = 3.715/8.064 + 5.5/9.6 * eta
    # 1.5PN
    fta3 = -3./4. * np.pi
    # 2PN
    fta4 = 9.275495/14.450688 + 2.84875/2.58048 * eta + 1.855/2.048 * eta*eta
    # 2.5PN
    fta5 = (3.8645/2.1504 - 6.5/25.6 * eta) * np.pi
    # 3PN
    fta6 = 83.1032450749357/5.7682522275840 - 5.3/4.0 * pi2 - 10.7/5.6 * GAMMA \
        + (-126.510089885/4.161798144 + 2.255/2.048 * pi2) * eta \
        + 1.54565/18.35008 * eta*eta - 1.179625/1.769472 * eta*eta*eta

    # 3.5PN
    fta7 = (1.88516689/1.73408256 + 4.88825/5.16096 *
            eta - 1.41769/5.16096 * eta*eta) * np.pi

    # 3PN log term
    ftal6 = -10.7/5.6

    # 2.5 PN log term
    ftal5 = np.log(theta)

    full = ftaN/theta5 * (fts1
                          + fta2*theta2
                          + fta3*theta3
                          + fta4*theta4
                          + (fta5*ftal5)*theta5
                          + (fta6 + ftal6*np.log(2.*theta))*theta6
                          + fta7*theta7)

    return m*(phi0 + full)



import tensorflow as tf
from scrinet.analysis import frequencyseries_batch as frequencyseries
from scrinet.analysis import timeseries_batch as timeseries


def fft(y):
    """ Fourier transform of y.
    This is fixed to assume:
        1. y is a real TimeSeries
        2. we use a real fft i.e. rfft
    Parameters
    ----------
    y : TimeSeries
        The input vector.
    Returns
    -------
    FrequencySeries:
        The (real) fourier transform of this (real) time series.
    """
    # need to assert either timeseries.TimeSeries timeseries_batch.TimeSeries
    # assert isinstance(y, timeseries.TimeSeries)
    y_data = tf.cast(y.data, tf.float32)
    f = tf.signal.rfft(y_data)
    # print(f)
    dt = tf.cast(y.delta_t, f.dtype)
    f = f * dt
    delta_f = 1.0/(tf.cast(y.delta_t, tf.float32) * tf.cast(tf.shape(y.data)[1], tf.float32))
    fs = frequencyseries.FrequencySeries(
        f, delta_f=delta_f, epoch=y.start_time)
    return fs


def ifft(ytilde, normalise=True):
    """ Inverse fourier transform of ytilde.
    This is fixed to assume:
        1. ytilde is a complex FrequencySeries
        2. we use a complex ifft i.e. ifft
    Parameters
    ----------
    ytilde : FrequencySeries
        The input vector.
    normalise : {bool, True}
        The output is always multiplied by the length of the ytilde data.
        If true then the output is also multiplied by delta_f * 2
    Returns
    -------
    TimeSeries:
        The inverse fourier transform of this frequency series.
    """
    # assert isinstance(ytilde, frequencyseries.FrequencySeries)
    # print(ytilde.data)
    t = tf.signal.ifft(ytilde.data)

    norm = tf.cast(tf.shape(ytilde.data)[1], tf.complex64)
    df = tf.cast(ytilde.delta_f, t.dtype)
    if normalise:
        norm *= df * 2

    norm = tf.cast(norm, tf.complex64)
    t = tf.cast(t, tf.complex64)
    t = t * norm

    delta_t = 1.0/(tf.cast(ytilde.delta_f, tf.float32) * tf.cast(tf.shape(ytilde.data)[1], tf.float32))
    return timeseries.TimeSeries(t, delta_t=delta_t, epoch=ytilde._epoch)



import tensorflow_probability as tfp

tfd = tfp.distributions
tfb = tfp.bijectors

def make_hmc_kernel(target_log_prob_fn,
                    num_burnin_steps,
                    step_size,
                    step_size_adapter,
                    num_leapfrog_steps,
                    target_accept_prob,
                    ):
    inner_kernel = tfp.mcmc.HamiltonianMonteCarlo(
        target_log_prob_fn=target_log_prob_fn,
        step_size=step_size,
        num_leapfrog_steps=num_leapfrog_steps)

    if step_size_adapter == 'simple':
        kernel = tfp.mcmc.SimpleStepSizeAdaptation(
            inner_kernel=inner_kernel,
            num_adaptation_steps=int(num_burnin_steps * 0.8),
            target_accept_prob=target_accept_prob)

    elif step_size_adapter == 'dual':
        kernel = tfp.mcmc.DualAveragingStepSizeAdaptation(
            inner_kernel=inner_kernel,
            num_adaptation_steps=int(num_burnin_steps * 0.8),
            target_accept_prob=target_accept_prob)
    else:
        raise ValueError

    return kernel


def make_nuts_kernel(target_log_prob_fn,
                     num_burnin_steps,
                     step_size,
                     step_size_adapter='simple',
                     max_tree_depth=8,
                     parallel_iterations=20,
                     target_accept_prob=0.7
                     ):
    inner_kernel = tfp.mcmc.NoUTurnSampler(
        target_log_prob_fn=target_log_prob_fn,
        step_size=step_size,
        max_tree_depth=max_tree_depth,
        parallel_iterations=parallel_iterations,
    )

    if step_size_adapter == 'simple':

        kernel = tfp.mcmc.SimpleStepSizeAdaptation(
            inner_kernel,
            num_adaptation_steps=int(num_burnin_steps * 0.8),
            step_size_setter_fn=lambda pkr, new_step_size: pkr._replace(
                step_size=new_step_size
            ),
            step_size_getter_fn=lambda pkr: pkr.step_size,
            log_accept_prob_getter_fn=lambda pkr: pkr.log_accept_ratio,
            target_accept_prob=target_accept_prob,
        )

    elif step_size_adapter == 'dual':
        kernel = tfp.mcmc.DualAveragingStepSizeAdaptation(
            inner_kernel,
            num_adaptation_steps=int(num_burnin_steps * 0.8),
            step_size_setter_fn=lambda pkr, new_step_size: pkr._replace(
                step_size=new_step_size
            ),
            step_size_getter_fn=lambda pkr: pkr.step_size,
            log_accept_prob_getter_fn=lambda pkr: pkr.log_accept_ratio,
            target_accept_prob=target_accept_prob,
        )
    else:
        ValueError

    return kernel


def make_rwmcmc_kernel(target_log_prob_fn):
    inner_kernel = tfp.mcmc.RandomWalkMetropolis(target_log_prob_fn,
                                                 new_state_fn=tfp.mcmc.random_walk_normal_fn(scale=0.01)
                                                 )
    kernel = tfp.mcmc.TransformedTransitionKernel(inner_kernel=inner_kernel,
                                                  bijector=[tfb.Identity(),
                                                            tfb.Identity(),
                                                            tfb.Identity(),
                                                            tfb.Identity()])
    return kernel




import tensorflow as tf
import tensorflow_probability as tfp
import numpy as  np

tfd = tfp.distributions

class Sine(tfp.distributions.Distribution):
    """
    Sine prior
    """

    def __init__(self,
                 minimum=0,
                 maximum=2 * np.pi,
                 ):
        self.minimum = minimum
        self.maximum = maximum

    def prob(self, value, name='prob', **kwargs):
        is_in_prior_range = tf.math.logical_and(tf.math.greater_equal(value, self.minimum),
                                                tf.math.less_equal(value, self.maximum))
        prob = tf.math.sin(value) / 2
        out_of_bounds = tf.zeros_like(prob, dtype=tf.float64)
        prob = tf.where(is_in_prior_range, prob, out_of_bounds)
        return prob

    def log_prob(self, value, name='log_prob', **kwargs):
        prob = self.prob(value)
        return tf.math.log(prob)


class Cosine(tfp.distributions.Distribution):
    """
    Cosine prior
    """

    def __init__(self,
                 minimum=-np.pi / 2,
                 maximum=np.pi / 2,
                 ):
        self.minimum = minimum
        self.maximum = maximum

    def prob(self, value, **kwargs):
        is_in_prior_range = tf.math.logical_and(tf.math.greater_equal(value, self.minimum),
                                                tf.math.less_equal(value, self.maximum))
        prob = tf.math.cos(value) / 2
        out_of_bounds = tf.zeros_like(prob, dtype=tf.float32)
        prob = tf.where(is_in_prior_range, prob, out_of_bounds)
        return prob

    def log_prob(self, value, **kwargs):
        prob = self.prob(value)
        return tf.math.log(prob)

    def sample(self, shape=(1,)):
        # this is not correct but probably okay for now
        return tfd.Uniform(self.minimum, self.maximum).sample(shape)




import tensorflow as tf
import numpy as np
import bilby
# import scrinet.sample.sample_helpers as nn_help
from typing import NamedTuple

H1 = bilby.gw.detector.get_empty_interferometer('H1')
L1 = bilby.gw.detector.get_empty_interferometer('L1')

h1_detector_tensor = tf.convert_to_tensor(H1.detector_tensor, dtype=tf.float64)
l1_detector_tensor = tf.convert_to_tensor(L1.detector_tensor, dtype=tf.float64)

detector_tensors = {'H1': tf.reshape(h1_detector_tensor, shape=(1, 3, 3)),
                    'L1': tf.reshape(l1_detector_tensor, shape=(1, 3, 3))}

H1_vertex = tf.convert_to_tensor(H1.geometry.vertex, dtype=tf.float32)
L1_vertex = tf.convert_to_tensor(L1.geometry.vertex, dtype=tf.float32)


detector_vertex = {'H1': tf.reshape(H1_vertex, shape=(3,)),
                    'L1': tf.reshape(L1_vertex, shape=(3,))}



def tf_gps_time_to_gmst(gps_time):
    """

    From bilby
    https://git.ligo.org/lscsoft/bilby/-/blob/master/bilby/core/utils.py#L322
    and turned into TF code

    NOTE THIS WILL NEED TO BE UPDATED TO TAKE INTO ACCOUNT THINGS LIKE
    CORRECTIONS TO THE SIDEREAL TIME SO THIS IS JUST A TOY MODEL...


    Convert gps time to Greenwich mean sidereal time in radians

    This method assumes a constant rotation rate of earth since 00:00:00, 1 Jan. 2000
    A correction has been applied to give the exact correct value for 00:00:00, 1 Jan. 2018
    Error accumulates at a rate of ~0.0001 radians/decade.

    Parameters
    -------
    gps_time: float
        gps time

    Returns
    -------
    float: Greenwich mean sidereal time in radians

    """
    #     LEAVING THIS IN AS A REMINDER THAT IF WE WANT TO DO ANY SERIOUS SCIENCE
    #     THEN WE SHOULD MAKE A PROPER TF CODE.
    #     warnings.warn(
    #         "Function gps_time_to_gmst deprecated, use "
    #         "lal.GreenwichMeanSiderealTime(time) instead",
    #         DeprecationWarning)
    omega_earth = 2 * np.pi * (1 / 365.2425 + 1) / 86400.
    gps_2000 = 630720013.
    gmst_2000 = (6 + 39. / 60 + 51.251406103947375 / 3600) * np.pi / 12
    correction_2018 = -0.00017782487379358614
    sidereal_time = omega_earth * (gps_time - gps_2000) + gmst_2000 + correction_2018
    gmst = tf.math.mod(tf.convert_to_tensor(sidereal_time, dtype=tf.float64), 2 * np.pi)
    return gmst



def tf_time_delay_geocentric(detector1, detector2, ra, dec, time):
    """
    #TODO STOLEN FROM https://git.ligo.org/lscsoft/bilby/-/blob/master/bilby/gw/utils.py
    Calculate time delay between two detectors in geocentric coordinates based on XLALArrivaTimeDiff in TimeDelay.c
    Parameters
    -------
    detector1: array_like
        Cartesian coordinate vector for the first detector in the geocentric frame
        generated by the Interferometer class as self.vertex.
    detector2: array_like
        Cartesian coordinate vector for the second detector in the geocentric frame.
        To get time delay from Earth center, use detector2 = np.array([0,0,0])
    ra: float
        Right ascension of the source in radians
    dec: float
        Declination of the source in radians
    time: float
        GPS time in the geocentric frame

    Returns
    -------
    float: Time delay between the two detectors in the geocentric frame

    """
    time = tf.convert_to_tensor(time, tf.float64)
    gmst = tf_gps_time_to_gmst(time)
    theta, phi = tf_ra_dec_to_theta_phi(ra, dec, gmst)
    omega = tf.convert_to_tensor([tf.math.sin(theta) * tf.math.cos(phi),
                                  tf.math.sin(theta) * tf.math.sin(phi),
                                  tf.math.cos(theta)])

    delta_d = detector2 - detector1
    omega = tf.cast(tf.transpose(omega), tf.float32)
    delta_d = tf.cast(delta_d, tf.float32)
    # https://git.ligo.org/lscsoft/bilby/-/blob/master/bilby/core/utils.py#L28
    speed_of_light = 299792458.0
    return tf.matmul(omega, tf.transpose(delta_d)) / speed_of_light


def tf_ra_dec_to_theta_phi(ra, dec, gmst):
    """ Convert from RA and DEC to polar coordinates on celestial sphere

    COPIED FROM BILBY - https://git.ligo.org/lscsoft/bilby/-/blob/master/bilby/core/utils.py
    Parameters
    -------
    ra: float
        right ascension in radians
    dec: float
        declination in radians
    gmst: float
        Greenwich mean sidereal time of arrival of the signal in radians

    Returns
    -------
    float: zenith angle in radians
    float: azimuthal angle in radians

    """
    phi = ra - gmst
    theta = np.pi / 2 - dec
    phi = tf.convert_to_tensor(phi, dtype=tf.float64)
    theta = tf.convert_to_tensor(theta, dtype=tf.float64)
    return theta, phi



def tf_theta_phi_to_ra_dec(theta, phi, gmst):
    ra = phi + gmst
    dec = np.pi / 2 - theta
    return ra, dec


def tf_get_polarization_tensor(ra, dec, time, psi, mode):
    """
    Calculate the polarization tensor for a given sky location and time

    See Nishizawa et al. (2009) arXiv:0903.0528 for definitions of the polarisation tensors.
    [u, v, w] represent the Earth-frame
    [m, n, omega] represent the wave-frame
    Note: there is a typo in the definition of the wave-frame in Nishizawa et al.
    Parameters
    -------
    ra: float
        right ascension in radians
    dec: float
        declination in radians
    time: float
        geocentric GPS time
    psi: float
        binary polarisation angle counter-clockwise about the direction of propagation
    mode: str
        polarisation mode

    Returns
    -------
    array_like: A 3x3 representation of the polarization_tensor for the specified mode.

    """
    time = tf.convert_to_tensor(time, tf.float64)
    gmst = tf_gps_time_to_gmst(time)
    theta, phi = tf_ra_dec_to_theta_phi(ra, dec, gmst)

    u = tf.concat([tf.math.cos(phi) * tf.math.cos(theta),
                   tf.math.cos(theta) * tf.math.sin(phi),
                   -tf.math.sin(theta)],
                  axis=1)

    v = tf.concat([-tf.math.sin(phi),
                   tf.math.cos(phi),
                   tf.zeros_like(phi)],
                  axis=1)

    psi = tf.convert_to_tensor(psi, dtype=tf.float64)

    m = -u * tf.math.sin(psi) - v * tf.math.cos(psi)
    n = -u * tf.math.cos(psi) + v * tf.math.sin(psi)

    if mode.lower() == 'plus':
        return tf.einsum('...i,...j->...ij', m, m) - tf.einsum('...i,...j->...ij', n, n)
    elif mode.lower() == 'cross':
        return tf.einsum('...i,...j->...ij', m, n) + tf.einsum('...i,...j->...ij', n, m)


def tf_get_antenna_response(detector, ra, dec, time, psi, mode):
    """
    https://git.ligo.org/lscsoft/bilby/-/blob/master/bilby/gw/detector/interferometer.py
    Todo find function
    :param detector:
    :param ra:
    :param dec:
    :param time:
    :param psi:
    :param mode:
    :return:
    """
    polarisation_tensor = tf_get_polarization_tensor(ra, dec, time, psi, mode)
    return tf.einsum('...ij,...ij->...', detector_tensors[detector], polarisation_tensor)


def make_detector_waveform_from_polarisations(detector,
                                              h_plus,
                                              h_cross,
                                              ra,
                                              dec,
                                              time,
                                              psi):
    """
    https://git.ligo.org/lscsoft/bilby/-/blob/master/bilby/gw/detector/interferometer.py
    get_detector_response

    :param detector:
    :param h_plus:
    :param h_cross:
    :param ra:
    :param dec:
    :param time:
    :param psi:
    :return:
    """
    det_plus_response = tf.cast(tf_get_antenna_response(detector, ra, dec, time, psi, 'plus'), tf.float32)
    det_cross_response = tf.cast(tf_get_antenna_response(detector, ra, dec, time, psi, 'cross'), tf.float32)

    det_plus_response = tf.reshape(det_plus_response, (h_cross.shape[0], -1))

    det_cross_response = tf.reshape(det_cross_response, (h_cross.shape[0], -1))

    detector_h = h_plus * det_plus_response + h_cross * det_cross_response
    return detector_h

# def tf_generate_surrogate_at_detector(params,
                                      # amp_model,
                                      # amp_basis,
                                      # phase_model,
                                      # phase_basis,
                                      # detector='H1'):
    # """
    # # Todo add docs
    # :param params:
    # :param amp_model:
    # :param amp_basis:
    # :param phase_model:
    # :param phase_basis:
    # :param detector:
    # :return:
    # """
    # # maybe use NamedTuple for this
    # intrinsic_params = params[:, 0:3]
    # extrinsic_params = params[:, 3:]

    # n_batch = params.shape[0]
    # # messy
    # ra = tf.reshape(extrinsic_params[:, 0], shape=(n_batch, -1))
    # dec = tf.reshape(extrinsic_params[:, 1], shape=(n_batch, -1))
    # time = tf.reshape(extrinsic_params[:, 2], shape=(n_batch, -1))
    # psi = tf.reshape(extrinsic_params[:, 3], shape=(n_batch, -1))
    # phase_shift = tf.reshape(extrinsic_params[:, 4], shape=(n_batch, -1))

    # sur_hp, sur_hc, _, _ = nn_help.generate_surrogate(
        # intrinsic_params,
        # amp_model=amp_model,
        # amp_basis=amp_basis,
        # phase_model=phase_model,
        # phase_basis=phase_basis,
        # phase_shift=phase_shift)

    # det_wvf = make_detector_waveform_from_polarisations(detector=detector,
                                                        # h_plus=sur_hp,
                                                        # h_cross=sur_hc,
                                                        # ra=ra,
                                                        # dec=dec,
                                                        # time=time,
                                                        # psi=psi)

    # return det_wvf



import numpy as np
import copy
import os

from scrinet.greedy import greedyrb
from scrinet.fits import poly
from scrinet.fits import gprfit
from scrinet.fits import nn

import george
from george import kernels


class Surrogate(object):
    def __init__(self, integration, basis_method='eim', output_dir=""):
        """
        basis_method {str}: default 'eim'. Can also be 'rb'
            Choose which method to use as the basis.
            Either 'eim' which is the empirical interpolation method
            or 'rb' which is the reduced basis method
        output_dir {str}: default ""
            directory where things will be saved.
            will default to current directory
        """
        self.integration = integration
        self.basis_method = basis_method
        self.output_dir = output_dir
        if self.output_dir:
            os.makedirs(f"{self.output_dir}", exist_ok=True)
        else:
            self.output_dir = "./"

        self.grb = greedyrb.GreedyReducedBasis(
            integration=self.integration
        )

        pass

    def build_seed_basis(self, ts, ts_coords):
        self.grb.build_seed_basis(ts=ts, ts_coords=ts_coords)

    def run_greedy_sweep(self, ts, ts_coords, greedy_tol=1e-4, verbose=False):
        self.grb.greedy_sweep(
            ts, ts_coords, greedy_tol=greedy_tol, verbose=verbose)

    def setup_eim(self):
        self.grb.setup_eim()

    def build_eim(self, ts):
        self.grb.build_eim(ts)

    def get_basis(self):
        if self.basis_method == 'eim':
            if hasattr(self.grb, 'eim') is False:
                self.setup_eim()
            basis = self.grb.eim.B
        elif self.basis_method == 'rb':
            basis = self.grb.basis
        return basis

    def compute_projection_coefficients(self, ts):
        """
        given an input training set (ts) compute projection
        coefficients using the current self.basis_method
        """

        if self.basis_method == 'eim':
            try:
                idxs = self.grb.eim.indices
            except AttributeError as error:
                print(
                    "Can't find EIM indicies. Perhaps you forgot to construct the EIM with self.grb.build_eim()", flush=True)
                raise
            alpha = np.transpose(ts)[idxs].T
        elif self.basis_method == 'rb':
            alpha = self.grb.compute_projection_coefficients_array(
                self.grb.basis, ts)

        return alpha

    def save_basis(self, filename):
        """
        saves the basis as a "filename.npy" file
        """
        basis = self.get_basis()
        np.save(os.path.join(self.output_dir, filename), basis)

    def fit_eim(self, X, y, maxdegs=None, max_deg_total=None, method='lr', epochs=1000, scaleX=False, scaleY=False, verbose=True, nn_outname="best.h5", validation_data=None, nn_verbose=0, batch_size=None, outname_prefix=""):
        """
        X: domain.shape = (N, M): N = number of greedy points, M = dimensionality.
            e.g., if non-spinning then M=1
        y: range.shape = (P, N) where P is the number of outputs
            EIM data or RB data
        method: default = 'lr' == linear polynomial basis regression
        epochs, scaleX, scaleY, nn_outname: only for method='nn'
        nn_outname: name of output h5 file. '.h5' get appended automatically
        """
        self.method = method
        self.scaleX = scaleX
        self.scaleY = scaleY

        if self.method == 'lr':
            if maxdegs is None:
                raise ValueError("maxdegs is None")
            if max_deg_total is None:
                raise ValueError("max_deg_total is None")

        if X.shape[1] > 1 and method == 'gpr':
            raise NotImplementedError(
                f"Dimension is: {X.shape[0]}. GPR only implemented for 1D at the moment.")

        fits = []

        if self.method == 'nn':
            fit = nn.RegressionANN()
            if verbose:
                print("\n====\ntraining NN\n====\n", flush=True)

            nn_outname = os.path.join(self.output_dir, nn_outname)
            self.history = fit.fit(
                X,
                y.T,
                input_dim=X.shape[1],
                noutput=y.shape[0],
                epochs=epochs,
                validation_data=validation_data,
                outname=nn_outname,
                scaleX=self.scaleX,
                scaleY=self.scaleY,
                verbose=nn_verbose,
                batch_size=batch_size
            )

            fit.load_model(nn_outname)
            fits.append(fit)

        if verbose and self.method == 'nn':
            self.plot_history(outname=outname_prefix)

        for i, ydata in enumerate(y):

            if self.method == 'lr':
                max_degs = poly.generate_degrees(
                    maxdegs=maxdegs, max_deg_total=max_deg_total)
                fit, _ = poly.findpolyfit(
                    X,
                    ydata,
                    max_degs=max_degs,
                    output_best=True,
                    verbose=False
                )
                fits.append(fit)
            elif self.method == 'gpr':
                #                 solver in [george.HODLRSolver, george.BasicSolver]
                solver = george.BasicSolver
                jitter = 1e-4
                yerr = np.zeros(ydata.shape) + jitter
                fit = gprfit.GPRFit(X, ydata, yerr=yerr)
                kernel = np.var(ydata) * \
                    kernels.ExpSquaredKernel(0.5, ndim=1, axes=0)
                fit.setup_kernel(kernel)

                fit.fit(solver=solver)
                fit.opt()
                fits.append(fit)

            if verbose:
                self.plot_fit(
                    X, ydata, fit, i, validation_data=validation_data, outname_prefix=outname_prefix)

        self.fits = fits

    def plot_history(self, outname=None):
        """
        outname {str}, default is None
            shoule be something like 'loss.png'
        """
        import matplotlib
        import matplotlib.pyplot as plt

        plt.figure()
        plt.plot(self.history.history['loss'], label='training loss')
        if 'val_loss' in self.history.history.keys():
            plt.plot(self.history.history['val_loss'], label='validation loss')
        plt.yscale('log')
        plt.ylabel('loss')
        plt.xlabel('epochs')
        plt.legend()
        plt.tight_layout()
        if outname:
            outname = os.path.join(self.output_dir, f"{outname}_loss.png")
            plt.savefig(outname)
        else:
            plt.show()
        plt.close()

    def plot_fit(self, X, ydata, fit, basis_idx, validation_data=None, outname_prefix=None):
        """
        outname_prefix {str}, default is None
            final outname will be f"{outname_prefix}_{basis_idx}.png"
        basis_idx {int}
            basis number to plot
        validation_data {2 tuple}, default is None
            (Xval, yval)
        """
        import matplotlib
        import matplotlib.pyplot as plt

        if self.method == 'lr':
            yhat = fit.predict(X)
        elif self.method == 'gpr':
            # haven't implemented uncertainty
            yhat, _ = fit.predict(X)
        elif self.method == 'nn':
            yhat = fit.predict(X).T[basis_idx]

        if X.shape[1] == 1:
            xhat = np.linspace(
                X.min(),
                X.max(),
                1000)
            if self.method == 'lr':
                yhat_cont = fit.predict(xhat.reshape(-1, 1))
                xhat = np.exp(xhat)
                x2 = np.exp(X.ravel())
            elif self.method == 'gpr':
                # haven't implemented uncertainty
                yhat_cont, _ = fit.predict(xhat.reshape(-1, 1))
                xhat = np.exp(xhat)
                x2 = np.exp(X.ravel())
            if self.method == 'nn':
                yhat_cont = fit.predict(xhat.reshape(-1, 1)).T[basis_idx]
                xhat = np.exp(xhat)
                x2 = np.exp(X.ravel())
            plt.figure()
            plt.title(f"basis number = {basis_idx}")

            if validation_data:
                xval = np.exp(validation_data[0])
                yval = validation_data[1][:, basis_idx]
                plt.scatter(xval, yval, label='validation')

            plt.plot(xhat, yhat_cont, label='extrap', c='C1')
            plt.scatter(x2, ydata, label='data', s=50, c='r', lw=2)
            plt.scatter(x2, yhat, label='fit', marker='x', s=100, lw=2, c='k')
            plt.legend()
            plt.tight_layout()
            if outname_prefix:
                outname = os.path.join(
                    self.output_dir, f"{outname_prefix}_{basis_idx}.png")
                plt.savefig(outname)
            else:
                plt.show()
            plt.close()
        else:
            plt.figure()
            plt.title(f"basis number = {basis_idx}")
            range_x = range(len(yhat))
            plt.scatter(range_x, ydata, label='data', s=50, c='r', lw=2)
            plt.scatter(range_x, yhat, label='fit',
                        marker='x', s=100, lw=2, c='k')
            plt.legend()
            plt.tight_layout()
            if outname_prefix:
                outname = os.path.join(
                    self.output_dir, f"{outname_prefix}_{basis_idx}.png")
                plt.savefig(outname)
            else:
                plt.show()
            plt.close()

    def predict(self, coords, log_index=0):
        """
        coords.shape = (1, M), M=dimension
        log_index: take the log of the log_index'th component.
            by default this is the 0th index.
            Use `log_index=None` to not modify the `coords`.
        """
        coords = np.array(coords)

        assert len(
            coords.shape) > 1, f"expected shape is (1, M) but got {coords.shape}"

        if log_index != None:
            coords[0, log_index] = np.log(coords[0, log_index])

        alpha = np.zeros(len(self.fits))

        if self.method == 'lr':
            for i, fit in enumerate(self.fits):
                alpha[i] = fit.predict(coords)
        elif self.method == 'gpr':
            for i, fit in enumerate(self.fits):
                # haven't implemented uncertainty
                alpha[i], _ = fit.predict(coords)
        elif self.method == 'nn':
            alpha = self.fits[0].predict(coords)[0]

        if self.basis_method == 'eim':
            basis = self.grb.eim.B
        elif self.basis_method == 'rb':
            basis = self.grb.basis

        return np.dot(alpha, basis)

    def validate_surrogate(self, vts, vts_coords, x=None, a=None, b=None):
        """
        vts: validation training set
            vts.shape() = (N, P)
                N = number of entries
                P = number of points e.g. length of time or frequency series
        vts_coords: validation training set coordinates / parameters
            vts_coords.shape() = (N, M).
                N = number of entries
                M = dimensionality
        x: np.ndarray of shape (P,). Default: None
        a, b: float/int. the start and end values of x for integration.
            Default: None
        """
        if self.method in ['lr', 'gpr', 'nn']:
            log_index = 0
        # elif self.method == 'nn':
        #     log_index = None

        if x is not None:
            if a is None:
                a = x[0]
            if b is None:
                b = x[-1]
            mask = (x >= a) & (x <= b)
            x = x[mask]
            integration = greedyrb.Riemann([a, b], len(x))
        else:
            integration = self.grb.integration
            # mask that is all True
            mask = np.ones(vts[0].shape).astype(np.bool)

        model_errors = np.zeros(len(vts))
        for i in range(len(vts)):
            yhat = self.predict([vts_coords[i]], log_index=log_index)[mask]
            y = vts[i][mask]
            # error = self.grb.integration.norm(y-yhat)**2
            error = integration.norm(y-yhat)**2
            model_errors[i] = error

        worst_error_index = np.argmax(model_errors)
        worst_case = vts_coords[worst_error_index]

        return model_errors, worst_case, worst_error_index



import tensorflow as tf
from scrinet.analysis.utils import tf_get_antenna_response, detector_vertex
from tensorflow_probability.python.math import interpolation as tfp_interp

import numpy as np
import arviz as az
import scrinet.analysis.timeseries_batch as ts
import scrinet.analysis.utils as ut
import lal
import phenom
from typing import NamedTuple

from scrinet.interfaces import lalutils


# @tf.function(experimental_compile=True, autograph=True, experimental_relax_shapes=True)

class WaveformParams(NamedTuple):
    """
    Object for passing waveform params to functions

    Attributes:
        distance: distance in metre - shape = N,1
        m_tot: total mass in Msun - shape = N,1
        mass_ratio: mass ratio (defined as q>1) - shape = N,1
        chi1: dimensionless spin for bh1  (defined over (-1, 1) - shape = N,1
        chi2: dimensionless spin for bh2 (defined over (-1, 1) - shape = N,1
        theta: sphharm polar - shape = N,1
        phi: sphharm azimuth - shape = N,1
        phase_shift: phase shift - shape = N,1

    """

    distance: tf.Tensor
    m_tot: tf.Tensor
    mass_ratio: tf.Tensor
    chi_1: tf.Tensor
    chi_2: tf.Tensor
    theta: tf.Tensor
    phi: tf.Tensor
    phase_shift: tf.Tensor


class DetectorParams(NamedTuple):
    """
    Object for passing detector response params to functions

    Attributes:
         ra: right ascension in radians - shape = N,1
         dec: declination in radians - shape = N,1
         psi: polarisation angle in radians - shape = N,1
         time: time (geocentrc maybe!?) - shape = N,1

    """

    ra: tf.Tensor
    dec: tf.Tensor
    psi: tf.Tensor
    time: tf.Tensor


class GwParamScalers(NamedTuple):
    """
    Object for passing scalers from sample space into physical space
    basically a normalisation so that parameters are defined over the same
    (scale to order of magnitude)

    Attributes:
        distance: distance in metre - shape = N,1
        m_tot: total mass in Msun - shape = N,1
        mass_ratio: mass ratio (defined as q>1) - shape = N,1
        chi1: dimensionless spin for bh1  (defined over (-1, 1) - shape = N,1
        chi2: dimensionless spin for bh2 (defined over (-1, 1) - shape = N,1
        theta: sphharm polar - shape = N,1
        phi: sphharm azimuth - shape = N,1
        phase_shift: phase shift - shape = N,1
        ra: right ascension in radians - shape = N,1
        dec: declination in radians - shape = N,1
        psi: polarisation angle in radians - shape = N,1
        time: time (geocentrc maybe!?) - shape = N,1

    """

    distance: float = 8. * 10 ** 3
    m_tot: float = 100.
    mass_ratio: float = 8.
    chi_1: float = 1.
    chi_2: float = 1.
    theta: float = 2 * np.pi
    phi: float = 2 * np.pi
    phase_shift: float = 1.
    ra: float = 2 * np.pi
    dec: float = np.pi / 2
    psi: float = 2 * np.pi
    time: float = 3.0


def amp_predict_hack(model, _input):
    """
    Function to re-create model.predict(input) but allows the graph to be built correctly
    (see https://github.com/tensorflow/tensorflow/issues/33997)

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to make a prediction
    :return:
    """
    x = _input
    x = x_scale_hack(model, x)
    pred = tf.convert_to_tensor(x)
    net_layers = model.model.layers
    for layer in net_layers:
        pred = layer(pred)
    return pred


# @tf.function(experimental_compile=True, autograph=True, experimental_relax_shapes=True)
def phase_predict_hack(model, _input):
    """
    Function to re-create model.predict(input) but allows the graph to be built correctly
    (see https://github.com/tensorflow/tensorflow/issues/33997)

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to make a prediction
    :return:
    """
    x = _input
    x = x_scale_hack(model, x)
    pred = tf.convert_to_tensor(x)
    net_layers = model.model.layers
    for layer in net_layers:
        pred = layer(pred)
    pred = y_inv_minmax_scale_hack(model, pred)
    return pred


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
def x_scale_hack(model, _input):
    """
    Function to re-create sklearn standard scaler for the input (as a tensorflow function)

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    x = _input
    # can hardcode 3 because surrogate is q, chi1, chi2
    x_dim = 3
    means = tf.constant([model.X_scalers[i].mean_[0]
                         for i in range(x_dim)])
    stds = tf.constant([model.X_scalers[i].scale_[0]
                        for i in range(x_dim)])
    means = tf.cast(means, tf.float32)
    stds = tf.cast(stds, tf.float32)
    x_scaled = (x - means) / stds
    x_scaled = tf.cast(x_scaled, tf.float32)
    return x_scaled


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
def get_model_y_std_scalers(model, _input):
    """
    Re-casts the sklearn output std scalers for each output basis

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    n_basis = 8
    std = [model.Y_scalers[i].scale_[0] for i in range(n_basis)]
    std = tf.stack(std, axis=0)
    std = tf.cast(std, tf.float32)
    return std


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
def get_model_y_mean_scalers(model, _input):
    """
    Re-casts the sklearn output mean scalers for each output basis

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    n_basis = 8
    mean = [model.Y_scalers[i].mean_[0] for i in range(n_basis)]
    mean = tf.stack(mean, axis=0)
    mean = tf.cast(mean, tf.float32)
    return mean


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
def y_inv_scale_hack(model, _input):
    """
    Re-creates the sklearn inverse standard scaler as a tensorflow function,

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    y_scaled = _input

    mean = get_model_y_mean_scalers(model, _input)

    std = get_model_y_std_scalers(model, _input)
    y = y_scaled * std + mean
    return y


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)


def get_model_y_minmax_scale_scalers(model, _input):
    """
    Re-casts the sklearn output scale scalers for each output basis

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    # I think this is always the same
    n_basis = 8
    scale = [model.Y_scalers[i].scale_[0] for i in range(n_basis)]
    scale = tf.stack(scale, axis=0)
    scale = tf.cast(scale, tf.float32)
    return scale


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
def get_model_y_minmax_min_scalers(model, _input):
    """
    Re-casts the sklearn output min scalers for each output basis

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    # I think this is always the same
    n_basis = 8
    mins = [model.Y_scalers[i].min_[0] for i in range(n_basis)]
    mins = tf.stack(mins, axis=0)
    mins = tf.cast(mins, tf.float32)
    return mins


# @tf.function(experimental_compile=True, experimental_relax_shapes=True)
def y_inv_minmax_scale_hack(model, _input):
    """
    Re-creates the sklearn inverse standard scaler as a tensorflow function,

    :param model: A trained scrinet Neural Network
    :param _input: The point at which you want to scale
    :return:
    """
    y_scaled = _input
    mins = get_model_y_minmax_min_scalers(model, _input)
    scale = get_model_y_minmax_scale_scalers(model, _input)
    y = (y_scaled - mins) / scale
    return y


# we need a function to tell us the time before merger for a particular frequency
# we can estimate this using PN
# from https://github.com/gwastro/pycbc/blob/master/pycbc/waveform/spa_tmplt.py
# non-spinning only
# removed the 'porder' and hardcoding max-order
# @tf.function(experimental_relax_shapes=True)
# @tf.function()


def tf_findchirp_chirptime(m1, m2, fLower):
    # variables used to compute chirp time
    m = m1 + m2
    eta = m1 * m2 / m / m
    c0T = c2T = c3T = c4T = c5T = c6T = c6LogT = c7T = 0.

    c7T = lal.PI * (14809.0 * eta * eta / 378.0 - 75703.0 *
                    eta / 756.0 - 15419335.0 / 127008.0)

    c6T = lal.GAMMA * 6848.0 / 105.0 - 10052469856691.0 / 23471078400.0 + \
          lal.PI * lal.PI * 128.0 / 3.0 + \
          eta * (3147553127.0 / 3048192.0 - lal.PI * lal.PI * 451.0 / 12.0) - \
          eta * eta * 15211.0 / 1728.0 + eta * eta * eta * 25565.0 / 1296.0 + \
          eta * eta * eta * 25565.0 / 1296.0 + tf.math.log(4.0) * 6848.0 / 105.0
    c6LogT = 6848.0 / 105.0

    c5T = 13.0 * lal.PI * eta / 3.0 - 7729.0 * lal.PI / 252.0

    c4T = 3058673.0 / 508032.0 + eta * (5429.0 / 504.0 + eta * 617.0 / 72.0)
    c3T = -32.0 * lal.PI / 5.0
    c2T = 743.0 / 252.0 + eta * 11.0 / 3.0
    c0T = 5.0 * m * lal.MTSUN_SI / (256.0 * eta)

    # This is the PN parameter v evaluated at the lower freq. cutoff
    xT = tf.pow(lal.PI * m * lal.MTSUN_SI * fLower, 1.0 / 3.0)
    x2T = xT * xT
    x3T = xT * x2T
    x4T = x2T * x2T
    x5T = x2T * x3T
    x6T = x3T * x3T
    x7T = x3T * x4T
    x8T = x4T * x4T

    # Computes the chirp time as tC = t(v_low)
    # tC = t(v_low) - t(v_upper) would be more
    # correct, but the difference is negligble.

    # This formula works for any PN order, because
    # higher order coeffs will be set to zero.
    return c0T * (1 + c2T * x2T + c3T * x3T + c4T * x4T + c5T * x5T + (
            c6T + c6LogT * tf.math.log(xT)) * x6T + c7T * x7T) / x8T


# @tf.function(experimental_compile=True)
# @tf.function()


def tf_chieffective(m1, m2, s1z, s2z):
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiralTaylorF2ReducedSpin.c#L46
    /* sanity checks */
    if (m1 <= 0) XLAL_ERROR(XLAL_EDOM);
    if (m2 <= 0) XLAL_ERROR(XLAL_EDOM);
    if (fabs(s1z) > 1) XLAL_ERROR(XLAL_EDOM);
    if (fabs(s2z) > 1) XLAL_ERROR(XLAL_EDOM);
    """
    m = m1 + m2
    eta = m1 * m2 / (m * m)
    delta = (m1 - m2) / m
    chi_s = (s1z + s2z) / 2.
    chi_a = (s1z - s2z) / 2.

    return chi_s * (1. - 76. * eta / 113.) + delta * chi_a


# @tf.function(experimental_compile=True)
# @tf.function()


def tf_XLALSimInspiralTaylorF2ReducedSpinChirpTime(
        m1,  # /**< mass of companion 1 (Msun) */
        m2,  # /**< mass of companion 2 (Msun) */
        chi,  # /**< dimensionless aligned-spin param */
        fStart  # /**< start GW frequency (Hz) */
):
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiralTaylorF2ReducedSpin.c#L307
#     if (fStart <= 0) XLAL_ERROR(XLAL_EDOM);
#     if (m1_SI <= 0) XLAL_ERROR(XLAL_EDOM);
#     if (m2_SI <= 0) XLAL_ERROR(XLAL_EDOM);
#     if (fabs(chi) > 1) XLAL_ERROR(XLAL_EDOM);


    returns tau (chirp time in seconds)

    """
    m = m1 + m2
    eta = m1 * m2 / (m * m)
    eta2 = eta * eta
    chi2 = chi * chi
    sigma0 = (-12769 * (-81. + 4. * eta)) / (16. * (-113. + 76. * eta) * (-113. + 76. * eta))
    gamma0 = (565 * (-146597. + 135856. * eta + 17136. * eta2)) / \
             (2268. * (-113. + 76. * eta))

    #     https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiralTaylorF2ReducedSpin.c#L31
    # # this is PI**2
    Pi_p2 = 9.8696044010893586188344909998761511

    v = tf.pow(lal.PI * m * lal.MTSUN_SI * fStart, 1. / 3.)
    vk = v  # /* v^k */
    #     REAL8 tk[8];  /* chirp time coefficients up to 3.5 PN */
    #     /* chirp time coefficients up to 3.5PN  */
    tk0 = (5. * m * lal.MTSUN_SI) / (256. * tf.pow(v, 8) * eta)
    tk1 = 0.
    tk2 = 2.9484126984126986 + (11 * eta) / 3.
    tk3 = (-32 * lal.PI) / 5. + (226. * chi) / 15.
    tk4 = 6.020630590199042 - 2 * sigma0 * chi2 + (5429 * eta) / 504. + (617 * eta2) / 72.
    tk5 = (3 * gamma0 * chi) / 5. - (7729 * lal.PI) / 252. + (13 * lal.PI * eta) / 3.
    tk6 = -428.291776175525 + (128 * Pi_p2) / 3. + (6848 * lal.GAMMA) / 105. + (3147553127 * eta) / 3.048192e6 - (
            451 * Pi_p2 * eta) / 12. - (15211 * eta2) / 1728. + (25565 * eta2 * eta) / 1296. + (
                  6848 * tf.math.log(4 * v)) / 105.
    tk7 = (-15419335 * lal.PI) / 127008. - (75703 * lal.PI * eta) / \
          756. + (14809 * lal.PI * eta2) / 378.

    vk2 = vk * v
    vk3 = vk2 * v
    vk4 = vk3 * v
    vk5 = vk4 * v
    vk6 = vk5 * v
    vk7 = vk6 * v

    tau = tk0 * (1 + tk2 * vk2 + tk3 * vk3 + tk4 * vk4 + tk5 * vk5 + tk6 * vk6 + tk7 * vk7)

    return tau


# @tf.function(experimental_compile=True)
# @tf.function()


def tf_get_surrogate_index_and_value_at_f_lower(m1, m2, s1z, s2z, flower):
    """
    returns the first index above the input flower
    and the value at that index
    """
    #     time_at_flower_sec = -tf_findchirp_chirptime(m1,m2,flower)
    chi = tf_chieffective(m1, m2, s1z, s2z)
    time_at_flower_sec = - \
        tf_XLALSimInspiralTaylorF2ReducedSpinChirpTime(m1, m2, chi, flower)
    time_at_flower_M = tf.cast(phenom.StoM(
        time_at_flower_sec, m1 + m2), tf.float64)
    # hardcoded to the final ann-sur model
    tstart_M = tf.cast(-20000., tf.float64)
    tend_M = 100.
    tnum = 40000

    # comment this out to support XLA compilation
    # tf.debugging.assert_greater_equal(
    #     time_at_flower_M,
    #     tstart_M,
    #     message=f"error: flower too low. flower = {flower}. time_at_flower_M = {time_at_flower_M}"
    # )

    surrogate_times = tf.linspace(tstart_M, tend_M, tnum)
    dt_M = surrogate_times[1] - surrogate_times[0]

    # Todo need to think about time, *0 below is bad
    target_indexs = tf.cast(
        tf.floor((20000. + time_at_flower_M) / dt_M), tf.int32)

    target_indexs = tf.clip_by_value(target_indexs,
                                     clip_value_max=tnum - 1,
                                     clip_value_min=0)
    vals = tf.gather(surrogate_times, target_indexs)

    return target_indexs, vals


# @tf.function


def y22_func(theta, phi):
    """
    https://lscsoft.docs.ligo.org/lalsuite/lal/_spherical_harmonics_8c_source.html#l00042
    """
    fac = tf.math.sqrt(5.0 / (64.0 * lal.PI)) * \
          (1.0 + tf.math.cos(theta)) * (1.0 + tf.math.cos(theta))

    m = tf.cast(2., tf.complex64)
    return tf.math.exp(1.j * m * tf.cast(phi, tf.complex64)) * tf.cast(fac, tf.complex64)


# @tf.function


def y2m2_func(theta, phi):
    """
    https://lscsoft.docs.ligo.org/lalsuite/lal/_spherical_harmonics_8c_source.html#l00042
    """
    fac = tf.math.sqrt(5.0 / (64.0 * lal.PI)) * \
          (1.0 - tf.math.cos(theta)) * (1.0 - tf.math.cos(theta))

    m = tf.cast(-2., tf.complex64)
    return tf.math.exp(1.j * m * tf.cast(phi, tf.complex64)) * tf.cast(fac, tf.complex64)


# @tf.function(experimental_compile=True)
# @tf.function()


def make_rect_window_tensor(M, start_end):
    B = []
    for se in start_end:
        B.append(tf.ones(se[1] - se[0]))

    C = [tf.pad(B[i], [[start_end[i][0], M - start_end[i][1]]])
         for i in range(len(start_end))]
    return tf.convert_to_tensor(C)


def apply_roll(tensor_in, shifts):
    """
    tensor_in: input tensor to roll - shape = N,M (N waveforms, M samples in time)
    shifts: tensor - shape = N,
        number of indicies to shift each row by

    returns: the rolled tensor

    """
    rolled = [tf.roll(tensor_in[i], shifts[i], axis=0)
              for i in range(tensor_in.shape[0])]
    return tf.convert_to_tensor(rolled)


# based on generate_surrogate in sample_helpers.py
# @tf.function(experimental_compile=True)
# @tf.function()


def generate_surrogate_hp_hc(
        waveform_params: WaveformParams,
        n_batch,
        f_lower,
        seglen,
        sample_rate,
        amp_model,
        amp_basis,
        phase_model,
        phase_basis
):
    """
    generate hplus and hcross. the shape is NxM
        N = number of waveforms
        M = seglen/samples_rate
    times start at 0s and extend for seglen seconds
    the returned data have the sample sample-rate and duration.
    they also start at the correct start frequency
    so waveforms with different parameters will have different lengths
    n_batch:
    f_lower, : shape = scalar
    seglen, : shape = scalar in seconds - duration of output times
    sample_rate, : shape = scalar in seconds

    :param theta: sphharm polar : shape = N,1
    :param phi: sphharm azimuth : shape = N,1

    :param x: The location in parameter space - currently (q, chi1, chi2) : shape = N,3
    :param amp_model: The pre-loaded amplitude nn generator
    :param amp_basis: The pre-loaded amplitude basis
    :param phase_model: The pre-loaded phase nn generator
    :param phase_basis: The pre-loaded phase basis
    :param phase_shift: phase shift (default: 0.) : shape = N,1

    :return:

    surr_times_sec: times in seconds: shape = (1,)
    hp_rolled: plus pol : shape = (N, M)
    hc_rolled: cross pol : shape = (N, M)
    -start_times: approximate time of merger : shape = (N, 1)


    """

    dt = 1. / sample_rate
    dt_M = phenom.StoM(dt, waveform_params.m_tot)

    # q = tf.transpose(tf.expand_dims(waveform_params.mass_ratio, 0))
    # s1z = tf.transpose(tf.expand_dims(waveform_params.chi_1, 0))
    # s2z = tf.transpose(tf.expand_dims(waveform_params.chi_2, 0))

    q = waveform_params.mass_ratio
    s1z = waveform_params.chi_1
    s2z = waveform_params.chi_2

    m1, m2 = phenom.m1_m2_M_q(waveform_params.m_tot, q)

    x = tf.concat([tf.math.log(waveform_params.mass_ratio),
                   waveform_params.chi_1,
                   waveform_params.chi_2], axis=1)

    x = tf.cast(x, tf.float32)

    amp_alpha = amp_predict_hack(amp_model, x)
    amp = tf.tensordot(amp_alpha, amp_basis, axes=1)
    phase_alpha = phase_predict_hack(phase_model, x)
    phase = tf.tensordot(phase_alpha, phase_basis, axes=1)
    phase_shift = waveform_params.phase_shift
    phase_shift = tf.cast(phase_shift, dtype=tf.float32)
    # phase = phase + phase_shift
    phase = phase
    amp_pre_fac = lalutils.td_amp_scale(waveform_params.m_tot,
                                        waveform_params.distance) * phenom.eta_from_q(q)

    amp_pre_fac = tf.cast(amp_pre_fac, tf.float32)

    amp = amp * amp_pre_fac

    idx, val = tf_get_surrogate_index_and_value_at_f_lower(
        m1, m2, s1z, s2z, f_lower - 5)

    original_idx = idx
    original_val = val  # i think this can be use to estimate the epoch
    val_idx = tf.argmin(idx)[0]
    # TODO is new_idx a constant? No this needs more thinkinf
    new_idx = tf.gather_nd(idx, [[val_idx, 0]])[0]
    # val = tf.cast(tf.gather_nd(val, [[val_idx, 0]])[0],
    #               dtype=tf.float32)
    # val = tf.cast(val[val_idx, 0], dtype=tf.float32)
    # these numbers are specific to the time axis of the surrogate
    x_ref_min = -20000.
    #    x_ref_min = val
    x_ref_max = 100.

    tfdt = tf.cast(dt_M, tf.float64)

    # TODO move all this time conversion stuff to its own function

    num_t_int = tf.cast(seglen / dt, tf.int32)
    num_T = tf.cast(num_t_int, tf.float64)

    times_M = tf.range(num_T) * tfdt - tf.abs(tf.cast(val, tf.float64))
    times_M_f32 = tf.cast(times_M, tf.float32)

    # x_ref_min = val * tf.ones(n_batch)
    x_ref_min = x_ref_min * tf.ones(n_batch)
    x_ref_max = x_ref_max * tf.ones(n_batch)

    # y_ref_gather =tf.stack(tf.gather_nd(amp, [:,new_idx:]))
    y_ref = tf.stack((amp, phase))
    # y_ref = tf.stack((amp[:, new_idx:], phase[:, new_idx:]))
    #
    # TODO Need to check if this actually produces the right waveform
    iy = tfp_interp.batch_interp_regular_1d_grid(
        x=times_M_f32,
        x_ref_min=x_ref_min,
        x_ref_max=x_ref_max,
        y_ref=y_ref,
        # y_ref=tf.stack((amp[:, new_idx:], phase[:, new_idx:])),
        #         fill_value=0 # comment this line out to work with tf.function.... not sure why....
    )

    new_amp = iy[0]
    new_phase = iy[1]

    phase = tf.cast(new_phase, tf.complex64)
    amp = tf.cast(new_amp, tf.complex64)

    theta = tf.cast(waveform_params.theta, tf.float32)
    phi = tf.cast(waveform_params.phi, tf.float32)

    h22 = amp * tf.math.exp(-1.j * phase)

    y22 = y22_func(theta, phi)
    y2m2 = y2m2_func(theta, phi)

    h = h22 * y22 + tf.math.conj(h22) * y2m2

    hplus = tf.math.real(h)
    hcross = tf.math.imag(h)

    surr_times_sec = phenom.MtoS(
        times_M,
        tf.cast(waveform_params.m_tot, tf.float64)
    )

    # negative becuase these times start at negative values
    epochs = -surr_times_sec[:, 0][:, tf.newaxis]
    surr_times_sec = surr_times_sec + epochs

    # finally here lets pad to the desired seglen
    # right now all waveforms have the same shape and sample-rate
    # so padding to desired seglen should be easy
    start_times = phenom.MtoS(original_val, tf.cast(waveform_params.m_tot,
                                                    tf.float64))
    epochs = tf.cast(epochs, tf.float64)
    dt = tf.cast(dt, tf.float64)

    extra_time = phenom.MtoS(150, tf.cast(waveform_params.m_tot, tf.float64))

    index_start_freq = tf.cast((epochs + start_times) / dt, tf.int32)

    index_peak_amp = tf.cast((epochs + extra_time) / dt, tf.int32)
    # need to work out a better way to estimate extra_time, should be related to hz
    # current implementation will be a problem for v high masses
    index_start_freq_buffer = tf.cast((epochs + start_times + 3 * extra_time) / dt, tf.int32)
    # ab = tf.transpose((index_start_freq_buffer, index_peak_amp))[0]
    ab = tf.transpose((index_start_freq, index_peak_amp))[0]
    # Can't XLA compile function below, only this one left
    window_zeros = tf.zeros((n_batch, num_t_int))
    # window = make_rect_window_tensor(num_t_int, [ab[0, i] for i in range(n_batch)])
    # window_ones = tf.ones((n_batch, num_t_int))

    ar = tf.range(num_t_int)
    cond = (ar >= ab[0][0]) & (ar < ab[0][1])
    window = tf.where(tf.math.logical_not(cond), window_zeros[0], 1.0)
    window = tf.reshape(window, (1, num_t_int))

    for i in range(1, n_batch):
        ar = tf.range(num_t_int)
        cond = (ar >= ab[i][0]) & (ar < ab[i][1])
        # cond = tf.math.logical_not(cond)
        window_next_bach = tf.where(tf.math.logical_not(cond), window_zeros[i], 1.0)
        window_next_bach = tf.reshape(window_next_bach, (1, num_t_int))

        window = tf.concat([window, window_next_bach], axis=0)

    cd = tf.transpose((index_start_freq,
                       index_start_freq_buffer))[0]
    cond_taper = (ar >= cd[0][0]) & (ar < cd[0][1])
    taper_ones = tf.ones((n_batch,
                           num_t_int))

    taper_len = tf.linspace(0., np.pi/2, num_t_int)
    P = tf.cast(num_t_int/index_start_freq_buffer[0][0], tf.float32)
    taper = tf.math.sin(P * taper_len) ** 2
    left_window = tf.where(tf.math.logical_not(cond_taper),
                           taper_ones[0],
                           taper)

    left_window = tf.reshape(left_window, (1, num_t_int))

    for i in range(1, n_batch):
        cond_taper = (ar >= cd[i][0]) & (ar < cd[i][1])
        P = tf.cast(num_t_int/index_start_freq_buffer[i][0], tf.float32)
        taper = tf.math.sin(P * taper_len) ** 2
        taper_next_batch = tf.where(tf.math.logical_not(cond_taper),
                               taper_ones[i],
                               taper)

        taper_next_batch = tf.reshape(taper_next_batch, (1, num_t_int))
        left_window = tf.concat([left_window, taper_next_batch], axis=0)

    # print(left_window)
    # test that window agrees with original implemtation
    # window = 1,
    # return amp
    # # uncomment this if you want to start the timeseries at 0
    # # otherwise below we roll such that the timeseries starts at the peak
    # #     hp_rolled = apply_roll(hplus*window, -tf.reshape(index_start_freq, (N,)))
    # #     hc_rolled = apply_roll(hcross*window, -tf.reshape(index_start_freq, (N,)))
    #
    index_peak_amp = tf.cast((epochs) / dt, tf.int32)

    hp_rolled = apply_roll(hplus * window * left_window, -tf.reshape(index_peak_amp, (n_batch,)))
    hc_rolled = apply_roll(hcross * window * left_window, -tf.reshape(index_peak_amp, (n_batch,)))

    # if we want to taper

    return surr_times_sec[0], hp_rolled, hc_rolled, -start_times


def make_detector_waveform_from_polarisations(detector,
                                              h_plus,
                                              h_cross,
                                              detector_params,
                                              n_batch,
                                              delta_t):
    """
    https://git.ligo.org/lscsoft/bilby/-/blob/master/bilby/gw/detector/interferometer.py
    get_detector_response

    :param detector:
    :param h_plus:
    :param h_cross:
    :param detector_params:
    :param delta_t
    :return:
    """
    ra = detector_params.ra
    dec = detector_params.dec
    psi = detector_params.psi
    time = detector_params.time

    num_waveforms = n_batch

    det_plus_response = tf.cast(tf_get_antenna_response(detector, ra, dec, time, psi, 'plus'), tf.float32)
    det_cross_response = tf.cast(tf_get_antenna_response(detector, ra, dec, time, psi, 'cross'), tf.float32)

    det_plus_response = tf.reshape(det_plus_response, (num_waveforms, 1))
    det_cross_response = tf.reshape(det_cross_response, (num_waveforms, 1))

    detector_h = h_plus * det_plus_response + h_cross * det_cross_response

    time_shift = ut.tf_time_delay_geocentric(detector_vertex[detector],
                                             tf.zeros((1,
                                                       3)),
                                             ra,
                                             dec,
                                             time)[0]

    # # Be careful to first subtract the two GPS times which are ~1e9 sec.
    # # And then add the time_shift which varies at ~1e-5 sec
    # # may be a hack below
    strain_data_start_time = 0.
    dt_geocent = time - strain_data_start_time
    dt = tf.cast(dt_geocent, tf.complex64) + tf.cast(time_shift, tf.complex64)

    detector_h_ts = ts.TimeSeries(detector_h,
                                  delta_t=delta_t)

    # TODO, need to check this is actually resolved
    detector_h_fs = detector_h_ts.to_frequencyseries()

    freqs = detector_h_fs.get_sample_frequencies()
    freqs = tf.cast(freqs, tf.complex64)
    fac = tf.math.exp(-1j * 2 * np.pi * dt * freqs)
    #
    detector_h_fs.data = detector_h_fs.data * fac

    #
    detector_h_ts = ts.TimeSeries(tf.math.real(detector_h_fs.to_timeseries().data), delta_t=delta_t)
    #
    return detector_h_ts.data


def tf_generate_surrogate_at_detector(waveform_params,
                                      detector_params,
                                      n_batch,
                                      f_lower,
                                      seglen,
                                      sample_rate,
                                      amp_model,
                                      amp_basis,
                                      phase_model,
                                      phase_basis,
                                      detector='H1'):
    """
    # Todo add docs
    :param params:
    :param amp_model:
    :param amp_basis:
    :param phase_model:
    :param phase_basis:
    :param detector:
    :return:
    """
    #         phase_shift=phase_shift)
    surr_times_sec, hplus, hcross, unrolled_peak_time = generate_surrogate_hp_hc(
        waveform_params=waveform_params,
        n_batch=n_batch,
        f_lower=f_lower,
        seglen=seglen,
        sample_rate=sample_rate,
        amp_model=amp_model,
        amp_basis=amp_basis,
        phase_model=phase_model,
        phase_basis=phase_basis
    )

    delta_t = surr_times_sec[1] - surr_times_sec[0]
    # return hplus, hcross
    hplus = ts.TimeSeries(hplus, delta_t=delta_t)
    hcross = ts.TimeSeries(hcross, delta_t=delta_t)
    # return hplus.data, hcross.data
    det_wvf = make_detector_waveform_from_polarisations(detector=detector,
                                                        h_plus=hplus.data,
                                                        h_cross=hcross.data,
                                                        detector_params=detector_params,
                                                        n_batch=n_batch,
                                                        delta_t=delta_t)
    #
    # # TODO returning a ts class and not a tensor may be problematic
    return det_wvf


def convert_tfp_chains_to_arviz_object(chains, param_names, varying_param_names=None):
    """
    Changes the output from sampling into an arviz object, this makes it easier to perform diagnostics
    and plots using arviz

    :param chains: The raw output from the tfp sampler
    :param param_names:

    :return:
    """
    if not varying_param_names:
        varying_param_names = param_names
    param_scalers = GwParamScalers()._asdict()

    samples_ar = np.array(chains.numpy())
    samples_right_shape = np.swapaxes(samples_ar, 0, 1)
    samples_dict = {key: param_scalers[key] * samples_right_shape[:, :, param_names.index(key)] for key in
                    varying_param_names}
    arv_samples = az.from_dict(samples_dict)
    return arv_samples


@tf.function()
def elbo_loss():
    q_samples = distribution.sample((256,))
    return -tf.reduce_mean(
        target_log_prob_fn(q_samples) - distribution.log_prob(q_samples)
    )


def make_mcmc_kernel_fn(target_log_prob_fn):
    return tfp.mcmc.RandomWalkMetropolis(
        target_log_prob_fn=target_log_prob_fn,
        new_state_fn=tfp.mcmc.random_walk_normal_fn(
            scale=step_size,
        ),
    )


def make_hmc_kernel_fn(target_log_prob_fn):
    return tfp.mcmc.HamiltonianMonteCarlo(
        target_log_prob_fn=target_log_prob_fn, step_size=step_size, num_leapfrog_steps=3
    )

# CODE TO TRAIN NEUTRA-LIKE BIJECTOR, MOVING HERE FOR NOW BUT NEED TO TURN INTO FUNCTION
# made = tfb.AutoregressiveNetwork(
#     params=2,
#     hidden_units=[
#         512,
#         512,
#         512,
#         512,
#     ],
#     activation="tanh",
#     dtype=tf.float64,
# )
#
# distribution = tfd.TransformedDistribution(
#     distribution=tfd.MultivariateNormalDiag(tf.zeros_like(p0[0]), tf.ones_like(p0[0])),
#     bijector=tfb.Chain([constraints, tfb.MaskedAutoregressiveFlow(made)]),
# )
#
# optimiser = tf.optimizers.Adam(lr=0.0001, amsgrad=True, clipnorm=0.0001)
#
# print("starting network training")
# starttime = datetime.datetime.now()
# n_training_steps = 5 * 10 ** 2
# losses = []
#
# for i in range(n_training_steps):
#     optimiser.minimise(elbo_loss, distribution.trainable_variables)
#     loss = elbo_loss()
#     losses.append(loss)
#     if i % 200 == 0:
#         print(f" finishing step {i}, loss = {np.mean(losses[-20:])}")
#     if np.isnan(loss):
#         raise ValueError
#
# endtime = datetime.datetime.now()
# duration = endtime - starttime
# print("Training complete")
# print(f"Training duration: {duration}")
#
# plt.figure(figsize=(14, 6))
# plt.plot(losses)
# # plt.yscale('log')
# plt.ylabel("loss")
# plt.xlabel("num training steps")
# plt.savefig(f"{image_dir}VI/bijector_loss.png")
# plt.close()
#
# smp = distribution.sample((5 * 10 ** 3, num_chains))
#
# vi_data = sample_helpers.convert_tfp_chains_to_arviz_object(
#     smp, param_names
# )
#
# print('VI summary')
# print(az.summary(vi_data))
# plt.figure()
# az.plot_pair(vi_data,
#              var_names=param_names,
#              kind='kde',
#              contour=True,
#              marginals=True)
#
# plt.savefig(f"{image_dir}VI/tmp_corner.png")
# plt.close()
#
# plt.figure()
# az.plot_posterior(vi_data,
#                   hdi_prob=0.9)
#
# plt.savefig(f"{image_dir}VI/post.png")
# plt.close()



"""
implements the greedy reduced basis
algorithm with
iterative modified gram-schmidt orthogonalisation

    - http://iacs-courses.seas.harvard.edu/courses/am205/slides/am205_lec08.pdf
    - https://www.math.uci.edu/~ttrogdon/105A/html/Lecture23.html
    - Algorithm from Hoffman, `Iterative Algorithms for Gram-Schmidt Orthogonalization`
    - Chad Galley rompy package
    - Blackman+ 2017
    - Field+ 2014
"""

import numpy as np
import rompy as rp


class Integration(object):
    """
    class for function integration
    based on Chad Galley's rompy
    https://github.com/Cyberface/rompy/blob/master/rompy/integrals.py
    """

    def __init__(self):
        pass

    def integral(self, f):
        return np.dot(self.weights, f)

    def dot(self, f, g):
        return np.dot(self.weights, f.conj()*g)

    def norm(self, f):
        return np.sqrt(np.dot(self.weights, f.conj()*f).real)


class Riemann(Integration):
    def __init__(self, interval, num):
        super().__init__()

        self.interval = interval
        self.num = num

        self.nodes, self.weights = \
            self.make_quadrature_rules(self.interval, self.num)

    def make_quadrature_rules(self, interval, num):
        a = interval[0]
        b = interval[1]
        nodes = np.linspace(a, b, num=num)
        weights = np.ones(num, dtype='double')
        weights = (b-a)/(num-1) * weights

        return nodes, weights


class Trapezoidal(Integration):
    def __init__(self, interval, num):
        super().__init__()

        self.interval = interval
        self.num = num

        self.nodes, self.weights = \
            self.make_quadrature_rules(self.interval, self.num)

    def make_quadrature_rules(self, interval, num):
        a = interval[0]
        b = interval[1]
        nodes = np.linspace(a, b, num=num)
        weights = np.ones(num, dtype='double')
        weights[0] = 0.5
        weights[-1] = 0.5
        weights = (b-a)/(num-1) * weights

        return nodes, weights


class GreedyReducedBasis(object):
    """
    based upon Chad Galley's rompy package
    """
#     def __init__(self, seed_training_space, seed_index, integration):

    def __init__(self, integration):
        """
        """
        self.integration = integration

        # 1. set initial basis to be the `seed_index`
        # vector from the `seed_training_space`
        # i.e. `seed_training_space[seed_index]`

        # 2. find new greedy point (GP)

        # 3. add new GP to basis

        # 4. validate basis

        pass

    def update_greedy_points(self, new_point):

        npoints = self.greedy_points.shape[0]
        try:
            ndim = self.greedy_points.shape[1]
            new = np.zeros(shape=(npoints+1, ndim))
        except:
            #             ndim = 1
            new = np.zeros(npoints+1)

        new[:-1] = self.greedy_points

        new[-1] = new_point
        self.greedy_points = new

        return

    def compute_projection_coefficients(self, basis, ts):
        """
        for a single new vector compute it's projection onto the basis
        """
        return np.array([self.integration.dot(ts, b) for b in basis])

    def compute_projection_coefficients_array(self, basis, ts):
        """
        loop over all vectors in the training set (ts)
        and compute their projection coefficients by
        computes their projection onto the basis

        returns:
            alphas: numpy.ndarray
                alphas.shape = (ts.shape[0], basis.shape[0])
        """

        n_vectors = ts.shape[0]
        n_basis = basis.shape[0]

        alpha = np.zeros(shape=(n_vectors, n_basis))

        for i in range(n_vectors):
            alpha[i] = self.compute_projection_coefficients(basis, ts[i])
        return alpha

    def compute_representation(self, alpha, basis):
        """
        if you have the basis coefficients then
        compute the representation with current basis
        """
        return np.dot(alpha, basis)

    def compute_representation_ts(self, ts):
        """
        giving a single training point compute the
        representation by the current basis
        """
        alpha = self.compute_projection_coefficients(self.basis, ts)
        return self.compute_representation(alpha, self.basis)

    def compute_greedy_error(self, ts):
        """
        computes greedy error for current basis and an input training set
        returns:
            greedy_errors: numpy.ndarray
                array of projection (greedy) errors
        """
        alphas = self.compute_projection_coefficients_array(self.basis, ts)

        n_vectors, n_dim = ts.shape
        greedy_errors = np.zeros(n_vectors)

        for i in range(n_vectors):
            rep = self.compute_representation(alphas[i], self.basis)

            greedy_errors[i] = self.integration.norm(ts[i] - rep)**2

        return greedy_errors

    def find_new_greedy_point(self, basis, ts):
        """
        using the current basis find the point
        in the training set that has the
        largest projection error

        returns:
            max_err_idx: int, index of maximum projection (greedy)
            error
        """

        greedy_errors = self.compute_greedy_error(ts)

        max_err_idx = np.argmax(greedy_errors)
        max_err = greedy_errors[max_err_idx]

        return max_err, max_err_idx

    def _add_basis(self, v, basis, tol, max_iter):
        """
        once we have found the
        vector in the training space
        with the largest greedy-error
        we want to add it to our basis.

        We do this by othogonalising it against
        the current basis using an
        iterative modified gram-schmidt algorithm (IMGS)
        """
        norm = self.integration.norm(v)
        e = v / norm

        flag, ctr = 0, 1
        while flag == 0:
            for b in basis:
                e -= b*self.integration.dot(b, e)
            new_norm = self.integration.norm(e)
            if new_norm / norm <= tol:
                norm = new_norm
                ctr += 1
                if ctr > max_iter:
                    msg = "WARNING max number of iter reached " \
                        + "basis may not be orthonormal."
                    print(msg, flush=True)
                    flag = 1
            else:
                flag = 1

        return e/new_norm

    def build_seed_basis(self, ts, ts_coords, tol=0.5, max_iter=3):
        """

        This builds an orthonormal basis out of the input
        training set `ts`.

        This is used as the `seed basis` so that things
        like deterministic or boundary points can be included.

        input:
            ts: numpy.ndarray
                The "training set" (ts).
                the rows are the individual vectors
                the columns are the components of each vector
                ts.shape[0] = number of vectors
                ts.shape[1] = number of components === dimensionality of each vector
            ts_coords: numpy.ndarray
                ts_coords.shape[0] = number of vectors
                ts_coords.shape[1] = number of parameters === dimensionality of parameter space
        returns:
            nothing

        sets the `basis` attribute
        """

        n_vectors, n_dim = ts.shape

        # save seed training set points to the greedy points
        self.greedy_points = ts_coords.copy()

        # allocated memory for basis
        basis = np.zeros(shape=(n_vectors, n_dim))

        # select first element of training set as the
        # initial basis vector
        # FIXME: should shuffle to pick a random
        # vector to be the first one.
        norm = self.integration.norm(ts[0])
        basis[0] = ts[0] / norm

        for i in range(1, n_vectors):
            basis[i] = self._add_basis(
                ts[i], basis[:i], tol=tol, max_iter=max_iter)

        self.basis = basis

    def greedy_sweep(self, ts, ts_coords, greedy_tol=1e-12, imgs_tol=0.5, imgs_max_iter=3, verbose=True):
        """
        for a given training set performs
        a number of `greedy_iter` stopping if
            1. tol is met - #TODO currently only this is implemented
            2. number of steps is met or
            3. if the training set has been exhausted
        """

        # list to store the indicies of the parameters in the current greedy
        # sweep to avoid adding the same vector twice
        self.indicies = []
        self.greedy_errors = []

        steps = len(ts)

        for step in range(steps):
            status = self.greedy_step(
                step, ts, ts_coords, greedy_tol, imgs_tol, imgs_max_iter, verbose)
            if status in ['tol', 'idx']:
                break
            else:
                continue

        self.nbasis = len(self.greedy_points)
        return

    def greedy_step(self, step, ts, ts_coords, greedy_tol, imgs_tol=0.5, imgs_max_iter=3, verbose=True):
        """
        perform an iteration of the greedy algorithm
        possibly adding a new vector to the basis
        """
        # index with worst representation error
        max_error, idx = self.find_new_greedy_point(self.basis, ts)
        if max_error <= greedy_tol:
            if verbose:
                print("tolerance reached. exiting", flush=True)
            return 'tol'

        self.greedy_errors.append(max_error)

        # check that potential new point is not already in the basis already
        # if it is then exit without adding it again
        # if it is not then add it.

        if step != 0:
            if idx in self.indicies:
                if verbose:
                    print(
                        "proposed new greedy point is already in basis. exiting", flush=True)
                return 'idx'

        if verbose:
            print(f"step = {step}, error = {max_error}", flush=True)

        self.indicies.append(idx)
        self.update_greedy_points(np.array(ts_coords[idx]))

        new_element = self._add_basis(
            ts[idx], self.basis, tol=imgs_tol, max_iter=imgs_max_iter)

        # extend basis array by one
        new_basis = np.zeros(
            shape=(self.basis.shape[0]+1, self.basis.shape[1]))
        new_basis[:-1] = self.basis
        new_basis[-1] = new_element

        self.basis = new_basis
        return

    def setup_eim(self):
        """
        instantiate instance of rp.EmpiricalInterpolant.
        Computes EIM basis
        """
        self.eim = rp.EmpiricalInterpolant(self.basis)

    def build_eim(self, eim_ts):
        """
        builds an empirical interpolant for the current basis
        input:
            eim_ts: numpy.ndarray
                training set generated at the self.greedy_points
        """
        if hasattr(self, 'eim') is False:
            self.setup_eim()

        self.eim.make_data(eim_ts)


class NewGreedyReducedBasis(rp.greedy._ReducedBasis, rp.greedy._IteratedModifiedGramSchmidt):
    def __init__(self, inner):
        """
        based upon Chad Galley's rompy package
        but I've added the ability to extend a basis through
        enrichment.

        inner: integration rule. Instance of rompy.Integration.
        """

        self.inner = inner
        rp.greedy._ReducedBasis.__init__(self, inner)
        rp.greedy._IteratedModifiedGramSchmidt.__init__(self, inner)

        # this is an L2 loss
        self.loss = self.proj_errors_from_alpha

    def load_basis(self, filename):
        """[summary]

        Args:
            filename (str): path to .npy file containing basis matrix
        """
        self.basis = np.load(filename)
        self.size = len(self.basis)

    def compute_training_set_norms(self, ts):
        """compute the norm of all elements in training set
        returns array of len(training_set)
        """
        return np.array([self.inner.norm(tt) for tt in ts])

    def iteration(self, step, errs, ts, ext_idx=None):
        """one iteration of reduced basis greedy algorithm

        returns 1 if index already selected
        else it doesn't return anything (returns None) and modifies
        self.indices, self.errors, self.basis and self.alphas
        """
        next_index = np.argmax(errs)
        if next_index in self.indices:
            print(
                ">> Warning(Index already selected): Exiting greedy algorithm", flush=True)
            return 1
        else:
            idx = step+1
            if ext_idx:
                ext_idx = ext_idx+1
            else:
                ext_idx = idx

#             print(f"idx: {idx} \t ext_idx: {ext_idx}")
            self.indices[idx] = next_index
            self.errors[idx] = errs[next_index]
            self.basis[ext_idx], _ = self.add_basis(
                ts[next_index], self.basis[:ext_idx])
            self.alphas[ext_idx] = self.alpha_arr(self.basis[ext_idx], ts)

    def make(self, ts, tol, seed_index, num=None, rel=False, verbose=True):
        """make initial basis

        rel: default is False. If true then will use relative greedy errors
        """

        # Npoints: length of training set
        # Nsamples: number of samples per training set element. e.g. number of time samples
        # Nbasis: max number of basis elements, defaults to len(ts)
        Npoints, Nsamples = ts.shape

        if num is None:
            Nbasis = len(ts)
        else:
            assert type(num) is int, "Expecting integer."
            assert num >= 0, "Requested number of basis vectors must be non-negative."
            Nbasis = num

        # allocate memory for arrays
        # could modify to work with complex data...
        self.errors = np.zeros(Nbasis, dtype=np.float64)
        # initialising indices to -1 to safe guard against
        # checking if next_index is already in self.indices
        self.indices = np.zeros(Nbasis, dtype=np.int32) - 1
        self.basis = np.zeros((Nbasis, Nsamples), dtype=np.float64)
        self.alphas = np.zeros((Nbasis, Npoints), dtype=np.float64)

        # pre-compute training set norms
        ts_norms = self.compute_training_set_norms(ts)

        # pick element to seed greedy algorithm
        self.indices[0] = seed_index
        self.basis[0] = ts[seed_index] / ts_norms[seed_index]
        self.alphas[0] = self.alpha_arr(self.basis[0], ts)
        # only true for L2 norm. But not sure why...
        self.errors[0] = np.max(ts_norms)**2

        if rel:
            tol *= self.errors[0]

        # loop over all elements in the training set except for the seed_index
        if verbose:
            print("\nStep \t Error", flush=True)
        step = 0
        flag = 0
        while step < Nbasis:
            if verbose:
                if rel:
                    print(
                        f"{step + 1} \t {self.errors[step]/self.errors[0]}", flush=True)
                else:
                    print(f"{step + 1} \t {self.errors[step]}", flush=True)
            # check if tolerance is met
            if self.errors[step] <= tol:
                if step == 0:
                    step += 1
                break
            # of if the number of basis vectors has been reached
            elif step == Nbasis-1:
                step += 1
                break
            # otherwise, add another point and basis vector
            else:
                # first compute the projection errors of the current basis
                # [:step+1] selects all alphas upto the current step.
                # this loss function could be trivially parallelised
                errs = self.loss(self.alphas[:step+1], norms=ts_norms)
                # then pass the errors into the `iteration` method
                # which will pick the case with the largest error and add it to the basis via IMGS.
                flag = self.iteration(step, errs, ts)

            # if previously selected index selected again then exit
            if flag == 1:
                step += 1
                break
            # otherwise, increment counter
            step += 1

        self.size = step
        self.indices = self.indices[:self.size]
        self.errors = self.errors[:self.size]
        self.basis = self.basis[:self.size]
        self.alphas = self.alphas[:self.size]

    def validate(self, ts):
        """validate basis.
        compute basis representation error against training set (ts) and record cases with error > tol.
        """
        # compute errors
        # this loop can be trivially parallelised
        validation_errors = np.array(
            [self.proj_error_from_basis(self.basis, tt) for tt in ts])
        return validation_errors

    def enrich(self, ts, tol, num=None, rel=False, verbose=True):
        """enrich basis (add training set (ts) to basis.)
        """
        # Npoints: length of training set
        # Nsamples: number of samples per training set element. e.g. number of time samples
        # Nbasis: max number of basis elements, defaults to len(ts)
        Npoints, Nsamples = ts.shape

        if num is None:
            Nbasis = len(ts)
        else:
            assert type(num) is int, "Expecting integer."
            assert num >= 0, "Requested number of basis vectors must be non-negative."
            Nbasis = num

        # allocate memory for arrays
        # could modify to work with complex data...
        self.errors = np.zeros(Nbasis, dtype=np.float64)
        # initialising indices to -1 to safe guard against
        # checking if next_index is already in self.indices
        self.indices = np.zeros(Nbasis, dtype=np.int32) - 1

        # project training set onto current basis
        current_alphas = np.array([self.alpha_arr(bb, ts)
                                   for bb in self.basis])

        # extend basis and alpha array storage
        basis_extension = np.zeros((Nbasis, Nsamples), dtype=np.float64)
        self.basis = np.row_stack((self.basis, basis_extension))

        alpha_extension = np.zeros((Nbasis, Npoints), dtype=np.float64)
        self.alphas = np.row_stack((current_alphas, alpha_extension))

#         print("extended arrays")
#         print(self.basis.shape)
#         print(self.alphas.shape)

        # pre-compute training set norms
        ts_norms = self.compute_training_set_norms(ts)

        # idx is the first index in the extended basis
        # (because python arrays start at indexing at zero)
        idx = self.size

#         self.errors[0] = np.max(self.loss(self.alphas[:idx+1], norms=ts_norms))

        # loop over max number of possible number of basis vectors to add.
        if verbose:
            print("\nStep \t Error", flush=True)
        for step in range(-1, Nbasis-1):
            # index for extended arrays
            ext_idx = idx + step
#             print(f"pre iter ext_idx: {ext_idx}")
            errs = self.loss(self.alphas[:ext_idx+1], norms=ts_norms)
            flag = self.iteration(step, errs, ts, ext_idx=ext_idx)
            if flag == 1:
                break

            # if first step
            # this is quite the messy bit of code...
            if step == -1:
                if rel:
                    tol *= self.errors[0]

            if self.errors[step+1] <= tol:
                print("tolerance met. exiting", flush=True)
                break
            if verbose:
                if rel:
                    print(
                        f"{step + 1} \t {self.errors[step+1]/self.errors[0]}", flush=True)
                else:
                    print(f"{step + 1} \t {self.errors[step+1]}", flush=True)

        self.size = ext_idx+2
        self.indices = self.indices[:step+2]
        self.errors = self.errors[:step+2]
        self.basis = self.basis[:self.size]
        self.alphas = self.alphas[:self.size]
